"""
Parses web-page contents.
"""

import calendar
import collections
import copy
import os
import re
import urllib.parse

import bs4
import dateutil.parser
import pandas as pd
import requests
from pyhelpers._cache import _print_failure_message
from pyhelpers.ops import confirmed, fake_requests_headers, update_dict_keys
from pyhelpers.store import load_data, save_data
from pyhelpers.text import find_similar_str

from .utils import cd_data, homepage_url, print_instance_connection_error


# == Preprocess contents ===========================================================================


def _parse_other_tags_in_td_contents(x):
    if isinstance(x, str):
        td_text = x.strip(' ')

    else:
        tag_name, td_text = x.name, x.text

        if tag_name == 'em':
            td_text = f'[{td_text}]'

        elif tag_name == 'q':
            td_text = f'"{td_text}"'

        elif tag_name in {'span', 'a'}:
            td_class, td_class_child = x.get('class'), x.find('span')

            if td_class == ['r']:
                if td_text == 'no CRS?':
                    td_text = f'\t\t / [{td_text}]'
                elif '\n ' in td_text:
                    td_text = ' '.join(
                        [f'\t\t{y}' if y.startswith('(') and y.endswith(')') else f' / [{y}]'
                         for y in td_text.split('\n ')])
                elif '(' not in td_text and ')' not in td_text:
                    td_text = f'\t\t / [{td_text}]'
                else:
                    td_text = f'\t\t{td_text}'

            elif not td_class and td_class_child:
                td_text = f'\t\t{td_text}'

    return td_text


def _prep_records(trs, ths, sep=' / '):
    ths_len = len(ths)

    records = []
    row_spanned = []

    for no, tr in enumerate(trs):
        data = []
        tds = tr.find_all(name='td')

        if len(tds) != ths_len:
            tds = tds[:ths_len]

        for td_no, td in enumerate(tds):
            if td.find('td'):
                text_ = [''] if td.find('a') is None else td.find('a').contents + ["\t\t / "]
            else:
                text_ = [_parse_other_tags_in_td_contents(x) for x in td.contents]
            # _move_element_to_end(text_, char='\t\t')
            text = ' '.join(sorted([x for x in text_ if x.strip(' ')], key=lambda x: '\t\t' in x))

            if sep:
                old_sep = re.compile(r'/?\r?\n')
                if len(re.findall(old_sep, text)) > 0:
                    text = re.sub(r'/?\r?\n', sep, text)

            if td.has_attr('rowspan'):
                row_spanned.append((no, int(td['rowspan']), td_no, text))

            data.append(text)

        records.append(data)

    return records, row_spanned


def _check_row_spanned(records, row_spanned):
    if row_spanned:
        records_ = records.copy()

        row_spanned_dict = collections.defaultdict(list)
        for i, *to_repeat in row_spanned:
            row_spanned_dict[i].append(to_repeat)

        for i, to_repeat in row_spanned_dict.items():
            for no_spans, idx, dat in to_repeat:
                for j in range(1, no_spans):
                    k = i + j
                    # if (dat in records[i]) and (dat != '\xa0'): # and (idx < len(records[i]) - 1):
                    #     idx += np.abs(records[i].index(dat) - idx, dtype='int64')
                    k_len = len(records_[k])
                    if k_len < len(records_[i]):
                        if k_len == idx:
                            records_[k].insert(idx, dat)
                        elif k_len > idx:
                            if records_[k][idx] != '':
                                records_[k].insert(idx, dat)
                            else:  # records[k][idx] == '':
                                records_[k][idx] = dat

    else:
        records_ = records

    return records_


def parse_tr(trs, ths, sep=' / ', as_dataframe=False):
    # noinspection PyUnresolvedReferences
    """
    Parses a list of HTML ``<tr>`` elements and extracts data from a table.

    This function processes the rows from a table (``<tr>`` tags) and assigns them to corresponding
    column headers (``<th>`` tags). It can return the data either as a list of lists or as a
    dataframe.

    See also [`PT-1 <https://stackoverflow.com/questions/28763891/>`_].

    :param trs: The content of ``<tr>`` tags from a web page table.
    :type trs: bs4.ResultSet | list
    :param ths: A list of column names (typically from ``<th>`` tags) for the table.
    :type ths: list | bs4.element.Tag
    :param sep: The separator to replace any separators found in the raw data;
        defaults to ``' / '``.
    :type sep: str | None
    :param as_dataframe: If ``True``, returns the data as a Pandas DataFrame; defaults to ``False``.
    :type as_dataframe: bool
    :return: A list of lists representing rows of the table,
        or a dataframe if ``as_dataframe`` is ``True``.
    :rtype: pandas.DataFrame | list[list]

    **Examples**::

        >>> from pyrcs.parser import parse_tr
        >>> import requests
        >>> import bs4
        >>> example_url = 'http://www.railwaycodes.org.uk/elrs/elra.shtm'
        >>> source = requests.get(example_url)
        >>> parsed_text = bs4.BeautifulSoup(source.content, 'html.parser')
        >>> ths_dat = [th.text for th in parsed_text.find_all('th')]
        >>> trs_dat = parsed_text.find_all(name='tr')
        >>> tables_list = parse_tr(trs=trs_dat, ths=ths_dat)  # returns a list of lists
        >>> type(tables_list)
        list
        >>> len(tables_list) // 100
        1
        >>> tables_list[0]
        ['AAL',
         'Ashendon and Aynho Line',
         '0.00 - 18.29',
         'Ashendon Junction',
         'Now NAJ3']
    """

    records, row_spanned = _prep_records(trs=trs, ths=ths, sep=sep)

    records = _check_row_spanned(records, row_spanned)

    if isinstance(ths, bs4.Tag):
        column_names = [th.get_text(strip=True) for th in ths.find_all('th')]
    elif all(isinstance(x, bs4.Tag) for x in ths):
        column_names = [th.get_text(strip=True) for th in ths]
    else:
        column_names = copy.copy(ths)

    n_columns = len(column_names)
    empty_rows = []

    for k, record in enumerate(records):
        n = n_columns - len(record)
        if n == n_columns:
            empty_rows.append(k)
        elif n > 0:
            record.extend(['\xa0'] * n)
        elif n < 0 and record[2] == '\xa0':
            del record[2]

    if len(empty_rows) > 0:
        for k in empty_rows:
            del records[k]

    if as_dataframe:
        records = pd.DataFrame(data=records, columns=column_names)

    return records


def parse_table(source, parser='html.parser', as_dataframe=False):
    """
    Parses HTML ``<tr>`` elements to create a table from the given source.

    This function extracts data from the ``<thead>`` and ``<tbody>`` elements of an HTML table
    and processes it into a list of lists (rows of the table) or a dataframe.

    :param source: The response object containing the HTML table from a requested URL.
    :type source: requests.Response
    :param parser: The parser to use for processing the HTML;
        options are ``'html.parser'`` (default), ``'html5lib'`` or ``'lxml'``.
    :type parser: str
    :param as_dataframe: If ``True``, the parsed data is returned as a dataframe.
        If ``False``, it returns a list of lists and column names; defaults to ``False``.
    :type as_dataframe: bool
    :return: A tuple containing a list of column names and a list of lists representing
        rows of the table; if ``as_dataframe=True``, returns a dataframe.
    :rtype: tuple[list, list] | pandas.DataFrame | list

    **Examples**::

        >>> from pyrcs.parser import parse_table
        >>> import requests
        >>> source_dat = requests.get(url='http://www.railwaycodes.org.uk/elrs/elra.shtm')
        >>> columns_dat, records_dat = parse_table(source_dat)
        >>> columns_dat
        ['ELR', 'Line name', 'Mileages', 'Datum', 'Notes']
        >>> type(records_dat)
        list
        >>> len(records_dat) // 100
        1
        >>> records_dat[0]
        ['AAL',
         'Ashendon and Aynho Line',
         '0.00 - 18.29',
         'Ashendon Junction',
         'Now NAJ3']
    """

    soup = bs4.BeautifulSoup(markup=source.content, features=parser)

    theads, tbodies = soup.find_all(name='thead'), soup.find_all(name='tbody')

    tables = []
    for thead, tbody in zip(theads, tbodies):
        ths = [th.get_text(strip=True) for th in thead.find_all(name='th')]
        trs = tbody.find_all(name='tr')

        if as_dataframe:
            dat = parse_tr(trs=trs, ths=ths, as_dataframe=as_dataframe)
        else:
            dat = ths, parse_tr(trs=trs, ths=ths)

        tables.append(dat)

    if len(tables) == 1:
        tables = tables[0]

    return tables, soup


def parse_date(str_date, as_date_type=False):
    """
    Parses a string representation of a date into a formatted date.

    This function attempts to parse a string date (even with slight errors or non-standard formats)
    into either a string in the "YYYY-MM-DD" format or a `datetime.date`_ object.

    .. _`datetime.date`: https://docs.python.org/3/library/datetime.html#datetime.date

    :param str_date: The date as a string, whose format can vary and may include month names
        or other elements.
    :type str_date: str
    :param as_date_type: If ``True``, returns the result as a `datetime.date` object;
        if ``False`` (default), returns the result as a formatted string.
    :type as_date_type: bool
    :return: The parsed date either as a string in "YYYY-MM-DD" format or as a date object.
    :rtype: str | datetime.date

    **Examples**::

        >>> from pyrcs.parser import parse_date
        >>> str_date_dat = '2020-01-01'
        >>> parse_date(str_date_dat)
        '2020-01-01'
        >>> str_date_dat = '2020-jan-01'
        >>> parse_date(str_date_dat)
        '2020-01-01'
        >>> parse_date(str_date_dat, as_date_type=True)
        datetime.date(2020, 1, 1)
    """

    try:
        temp_date = dateutil.parser.parse(timestr=str_date, fuzzy=True)
        # or, temp_date = datetime.datetime.strptime(str_date[12:], '%d %B %Y')

    except (TypeError, calendar.IllegalMonthError):
        month_name = find_similar_str(str_date, lookup_list=calendar.month_name)
        err_month_ = find_similar_str(month_name, lookup_list=str_date.split(' '))

        temp_date = dateutil.parser.parse(
            timestr=str_date.replace(err_month_, month_name), fuzzy=True)

    parsed_date = temp_date.date() if as_date_type else str(temp_date.date())

    return parsed_date


# == Extract information ===========================================================================


def _clean_key(k_text):
    return k_text.replace("–", "-").strip("()").removesuffix(".shtml").removesuffix(".shtm")


def _parse_dd_or_dt(dd_or_dt):
    """
    Extracts text and href attributes from dt or dd elements.
    """

    dd_or_dt_contents = dd_or_dt.contents

    if len(dd_or_dt_contents) == 1:
        content = dd_or_dt_contents[0]
        if isinstance(content, str):
            text, href = content, None
        else:
            text = content.get_text(strip=True)
            href = content.get(key='href') if content.name == 'a' else None

    # Case 2: Two elements (text with a hyperlink reference)
    else:
        a_href, text = dd_or_dt_contents
        if not isinstance(text, str):
            text, a_href = dd_or_dt_contents
        # if re.search(r'\((.*?)\)', text) and text[1].islower():
        #     text = f'{text[1].upper()}{text[2:-1]}'
        href = a_href.find('a').get('href')

    return _clean_key(text), href


def _get_site_map_h3_dl_dt_dds(h3_dl_dt, next_dd=None):
    if next_dd is None:
        next_dd = h3_dl_dt.find_next('dd')

    prev_dt = next_dd.find_previous('dt')

    h3_dl_dt_dds = {}
    while prev_dt == h3_dl_dt:
        next_dd_sub_dl_ = next_dd.find('dl')

        if next_dd_sub_dl_:
            sub_dts = next_dd_sub_dl_.find_all('dt')

            for sub_dt in sub_dts:
                sub_dt_text, _ = _parse_dd_or_dt(sub_dt)
                sub_dt_dds = sub_dt.find_next_siblings('dd')
                sub_dt_dds_dict = _get_site_map_sub_dl(h3_dl_dts=sub_dt_dds)

                h3_dl_dt_dds.update({_clean_key(sub_dt_text): sub_dt_dds_dict})

        else:
            a = next_dd.find('a')
            text, href = _clean_key(a.get_text(strip=True)), a.get(key='href')
            h3_dl_dt_dds.update({text: urllib.parse.urljoin(homepage_url(), href)})

        try:
            next_dd = next_dd.find_next_sibling('dd')
            prev_dt = next_dd.find_previous_sibling('dt')
        except AttributeError:
            break

    return h3_dl_dt_dds


def _get_site_map_sub_dl(h3_dl_dts):
    """
    Recursively processes nested dl/dt/dd elements to build a structured dictionary.
    """

    h3_dl_dt_dd_dict = {}

    for h3_dl_dt in h3_dl_dts:
        dt_text_, dt_href = _parse_dd_or_dt(dd_or_dt=h3_dl_dt)
        dt_text = _clean_key(dt_text_)

        if dt_href:
            h3_dl_dt_dd_dict.update({dt_text: urllib.parse.urljoin(homepage_url(), dt_href)})

        else:
            next_dd = h3_dl_dt.find_next('dd')
            next_dd_sub_dl = next_dd.find('dl')

            if next_dd_sub_dl:
                # next_dd_sub_dl_dts = next_dd_sub_dl.find_all(name='dt')
                next_dd_sub_dl_dts = [
                    dt for dt in next_dd.find_all('dt') if dt.has_attr('class')]
                h3_dl_dt_dd_dict.update({dt_text: _get_site_map_sub_dl(next_dd_sub_dl_dts)})

            else:
                h3_dl_dt_dds = _get_site_map_h3_dl_dt_dds(h3_dl_dt=h3_dl_dt, next_dd=next_dd)
                h3_dl_dt_dd_dict.update({dt_text: h3_dl_dt_dds})

    return h3_dl_dt_dd_dict


def _get_site_map(source, parser='html.parser'):
    """
    Parses the site map from the given HTML source and returns a structured dictionary.
    """

    soup = bs4.BeautifulSoup(markup=source.content, features=parser)
    site_map = {}

    h3s = soup.find_all('h3', attrs={"class": "site"})

    for h3 in h3s:
        h3_title = h3.get_text(strip=True)
        h3_dl_dts = h3.find_next('dl').find_all('dt')  # h3 > dl > dt

        if len(h3_dl_dts) == 1:
            dd_dict = {}  # h3 > dl > dt > dd

            h3_dl_dt = h3_dl_dts[0]
            h3_dl_dt_text = h3_dl_dt.get_text(strip=True)

            if h3_dl_dt_text == '':
                for dd in h3_dl_dt.find_next_siblings('dd'):
                    text, href = _parse_dd_or_dt(dd)
                    dd_dict.update({_clean_key(text): urllib.parse.urljoin(homepage_url(), href)})

        else:
            dd_dict = _get_site_map_sub_dl(h3_dl_dts=h3_dl_dts)

        site_map.update({_clean_key(h3_title): dd_dict})

    # noinspection SpellCheckingInspection
    site_map = update_dict_keys(
        site_map, replacements={"(all ogher languages)": "(all other languages)"})

    return site_map


def get_site_map(update=False, confirmation_required=True, verbose=False, raise_error=True):
    # noinspection PyShadowingNames
    """
    Gets the `site map <http://www.railwaycodes.org.uk/misc/sitemap.shtm>`_.

    :param update: Whether to check for updates to the package data; defaults to ``False``.
    :type update: bool
    :param confirmation_required: Whether user confirmation is required before proceeding;
        defaults to ``True``.
    :type confirmation_required: bool
    :param verbose: Whether to print relevant information to the console; defaults to ``False``.
    :type verbose: bool | int
    :param raise_error: Whether to raise the provided exception;
        if ``raise_error=False``, the error will be suppressed; defaults to ``True``.
    :type raise_error: bool
    :return: A dictionary containing the data of site map.
    :rtype: dict | None

    **Examples**::

        >>> from pyrcs.parser import get_site_map
        >>> site_map = get_site_map()
        >>> type(site_map)
        dict
        >>> list(site_map.keys())
        ['Home',
         'Line data',
         'Other assets',
         '"Legal/financial" lists',
         'Miscellaneous']
        >>> site_map['Home']
        {'index': 'http://www.railwaycodes.org.uk/index.shtml'}
    """

    path_to_file = cd_data("site-map.json", mkdir=True)

    if os.path.isfile(path_to_file) and not update:
        return load_data(path_to_file)

    else:
        if confirmed("To collect the site map\n?", confirmation_required=confirmation_required):
            if verbose in {True, 1}:
                print("Updating the package data", end=" ... ")

            try:
                url = urllib.parse.urljoin(homepage_url(), '/misc/sitemap.shtm')
                source = requests.get(url=url, headers=fake_requests_headers())
                source.raise_for_status()
            except Exception as e:
                print_instance_connection_error(
                    update=update, verbose=True if update else verbose, e=e,
                    raise_error=raise_error)
                return None

            try:
                site_map = _get_site_map(source=source)

                if verbose in {True, 1}:
                    print("Done.")

                if site_map:
                    save_data(site_map, path_to_file, indent=4, verbose=(verbose == 2 or False))

                return site_map

            except Exception as e:
                _print_failure_message(
                    e, prefix="Failed. Error:", verbose=verbose, raise_error=raise_error)

        else:
            if verbose in {True, 1}:
                print("Cancelled. ")
            # site_map = load_data(path_to_file)


def _get_last_updated_date(soup, parsed=True, as_date_type=False):
    # Find 'Last update date'
    update_tag = soup.find(name='p', attrs={'class': 'update'})

    if update_tag is not None:
        last_updated_date = update_tag.get_text(strip=True)
        # Decide whether to convert the date's format
        if parsed:
            # Convert the date to "yyyy-mm-dd" format
            last_updated_date = parse_date(str_date=last_updated_date, as_date_type=as_date_type)

    else:
        last_updated_date = None

    return last_updated_date


def get_last_updated_date(url, parsed=True, as_date_type=False, verbose=False, raise_error=True):
    # noinspection PyShadowingNames
    """
    Gets the last update date of a specified web page.

    This function extracts the date when the given web page was last updated.
    The date can be returned as a string or a date object.

    :param url: The URL of the web page for which the last update date is requested.
    :type url: str
    :param parsed: Whether to reformat the date into a standardized format (``YYYY-MM-DD``);
        defaults to ``True``.
    :type parsed: bool
    :param as_date_type: If ``True``, the date is returned as a `datetime.date`_ object;
        if ``False`` (default), it's returned as a string.
    :type as_date_type: bool
    :param verbose: Whether to print relevant information to the console; defaults to ``False``.
    :type verbose: bool | int
    :param raise_error: Whether to raise the provided exception;
        if ``raise_error=False``, the error will be suppressed; defaults to ``True``.
    :type raise_error: bool
    :return: The last update date of the specified web page,
        or ``None`` if this information is not available on the web page.
    :rtype: str | datetime.date | None

    .. _`datetime.date`: https://docs.python.org/3/library/datetime.html#datetime.date

    **Examples**::

        >>> from pyrcs.parser import get_last_updated_date
        >>> url = 'http://www.railwaycodes.org.uk/crs/CRSa.shtm'
        >>> last_upd_date = get_last_updated_date(url=url, parsed=True, as_date_type=False)
        >>> type(last_upd_date)
        str
        >>> last_upd_date = get_last_updated_date(url=url, parsed=True, as_date_type=True)
        >>> type(last_upd_date)
        datetime.date
        >>> url = 'http://www.railwaycodes.org.uk/linedatamenu.shtm'
        >>> last_upd_date = get_last_updated_date(url=url, verbose=True)
        Information of the last update date not available.
    """

    try:  # Request to get connected to the given url
        source = requests.get(url=url, headers=fake_requests_headers())
        source.raise_for_status()
    except Exception as e:
        _print_failure_message(e, verbose=verbose, raise_error=raise_error)

    else:
        # Parse the text scraped from the requested web page
        soup = bs4.BeautifulSoup(markup=source.content, features='html.parser')

        last_updated_date = _get_last_updated_date(
            soup=soup, parsed=parsed, as_date_type=as_date_type)

        if last_updated_date is None and verbose:
            print('Information of the last update date not available.')

        return last_updated_date


def get_financial_year(date):
    """
    Gets the financial year of a given date.

    The financial year runs from 1st April to 31st March of the following year.
    This function takes a date and determines the financial year it falls into.

    :param date: The date for which the financial year is to be determined.
    :type date: datetime.datetime
    :return: The financial year of the given ``date``.
    :rtype: int

    **Examples**::

        >>> from pyrcs.parser import get_financial_year
        >>> import datetime
        >>> financial_year = get_financial_year(date=datetime.datetime(2021, 3, 31))
        >>> financial_year
        2020
    """

    financial_date = date + pd.DateOffset(months=-3)

    return financial_date.year


def _parse_introduction(source, delimiter='\n'):
    soup = bs4.BeautifulSoup(markup=source.content, features='html.parser')

    intro_h3 = [h3 for h3 in soup.find_all('h3') if h3.get_text(strip=True).startswith('Intro')][0]

    p = intro_h3.find_next(name='p')
    prev_h3, prev_h4 = p.find_previous(name='h3'), p.find_previous(name='h4')

    intro_paras = []
    while prev_h3 == intro_h3 and prev_h4 is None:
        para_text = p.text.replace('  ', ' ')
        intro_paras.append(para_text)

        p = p.find_next(name='p')
        prev_h3, prev_h4 = p.find_previous(name='h3'), p.find_previous(name='h4')

    introduction = delimiter.join(intro_paras)

    return introduction


def get_introduction(url, delimiter='\n', update=False, verbose=False, raise_error=False):
    """
    Gets the introduction section of a specified web page.

    This function scrapes the introduction text from the given URL, typically used to
    summarise data clusters.

    :param url: The URL of the web page (usually the main page of a data cluster).
    :type url: str
    :param delimiter: The delimiter used to separate paragraphs in the returned content;
        defaults to ``'\\n'`` (newline).
    :type delimiter: str
    :param update: Whether to check for updates to the package data; defaults to ``False``.
    :type update: bool
    :param verbose: Whether to print relevant information to the console; defaults to ``False``.
    :type verbose: bool | int
    :param raise_error: Whether to raise the provided exception;
        if ``raise_error=False`` (default), the error will be suppressed.
    :type raise_error: bool
    :return: The introductory text from the web page, formatted with the specified delimiter.
    :rtype: str

    **Examples**::

        >>> from pyrcs.parser import get_introduction
        >>> bridges_url = 'http://www.railwaycodes.org.uk/bridges/bridges0.shtm'
        >>> intro_text = get_introduction(url=bridges_url)
        >>> intro_text
        "There are thousands of bridges over and under the railway system. These pages attempt to...
    """

    intro_filename = '-'.join(x for x in urllib.parse.urlparse(url).path.replace(
        '.shtm', '.pkl').split('/') if x)
    path_to_file = cd_data("introduction", intro_filename, mkdir=True)

    if os.path.isfile(path_to_file) and not update:
        return load_data(path_to_file)

    try:
        source = requests.get(url=url, headers=fake_requests_headers())
    except Exception as e:
        print_instance_connection_error(
            update=update, verbose=True if update else verbose, e=e, raise_error=raise_error)
        return None

    try:
        introduction = _parse_introduction(source=source, delimiter=delimiter)

        if introduction:
            save_data(introduction, path_to_file=path_to_file, verbose=verbose)

        return introduction

    except Exception as e:
        _print_failure_message(e=e, prefix="Failed.", verbose=verbose, raise_error=raise_error)


def _parse_catalogue(source, url):
    """
    Extracts a catalogue of links from the provided ``BeautifulSoup4`` object.

    :param source: HTML content.
    :param url: Base (page) URL to resolve relative links.
    :return: Typically, a dictionary mapping link text to absolute URLs.
    """

    soup = bs4.BeautifulSoup(markup=source.content, features='html.parser')

    # Try to find the primary container, fallback to alternative
    cold_soup = soup.find(name='div', attrs={'class': 'fixed'}) or soup.find(name='h1')

    # Extract anchor tags (if fallback is used, get all following <a> tags)
    links = cold_soup.find_all('a') if cold_soup else []

    if len(links) > 0:
        catalogue = {
            a.text.replace('\xa0', ' ').strip(): urllib.parse.urljoin(url, a.get('href'))
            for a in links}
    else:
        catalogue = None

    return catalogue


def get_catalogue(url, update=False, json_it=True, verbose=False, raise_error=False):
    """
    Gets the catalogue of items from the main page of a data cluster.

    This function scrapes a catalogue of entries (typically hyperlinks) from a specified URL.
    It offers the option to save the catalogue as a JSON file.

    :param url: The URL of the main page of a data cluster.
    :type url: str
    :param update: Whether to check for updates to the package data; defaults to ``False``.
    :type update: bool
    :param json_it: Whether to save the catalogue as a JSON file; defaults to ``True``.
    :type json_it: bool
    :param verbose: Whether to print relevant information to the console; defaults to ``False``.
    :type verbose: bool | int
    :param raise_error: Whether to raise the provided exception;
        if ``raise_error=False`` (default), the error will be suppressed.
    :type raise_error: bool
    :return: The catalogue in the form of a dictionary, where keys are entry titles and
        values are URLs, or ``None`` if the operation is unsuccessful.
    :rtype: dict | None

    **Examples**::

        >>> from pyrcs.parser import get_catalogue
        >>> elr_cat = get_catalogue(url='http://www.railwaycodes.org.uk/elrs/elr0.shtm')
        >>> type(elr_cat)
        dict
        >>> list(elr_cat.keys())[:5]
        ['Introduction', 'A', 'B', 'C', 'D']
        >>> list(elr_cat.keys())[-5:]
        ['Lines without codes',
         'ELR/LOR converter',
         'LUL system',
         'DLR system',
         'Canals']
        >>> location_code_cat = get_catalogue(url='http://www.railwaycodes.org.uk/crs/crs0.shtm')
        >>> type(location_code_cat)
        dict
        >>> list(location_code_cat.keys())[:5]
        ['Introduction', 'A', 'B', 'C', 'D']
        >>> list(location_code_cat.keys())[-5:]
        ['W', 'X', 'Y', 'Z', 'Other systems']
    """

    cat_filename = '-'.join(x for x in urllib.parse.urlparse(url).path.replace(
        '.shtm', '.json').split('/') if x)
    path_to_file = cd_data("catalogue", cat_filename, mkdir=True)

    if os.path.isfile(path_to_file) and not update:
        return load_data(path_to_file)

    try:
        source = requests.get(url=url, headers=fake_requests_headers())
        source.raise_for_status()
    except Exception as e:
        _print_failure_message(e=e, verbose=verbose, raise_error=raise_error)
        return None

    try:
        catalogue = _parse_catalogue(source=source, url=url)

        if catalogue and json_it:
            save_data(catalogue, path_to_file=path_to_file, verbose=verbose, indent=4)

        return catalogue

    except Exception as e:
        _print_failure_message(e=e, prefix="Failed.", verbose=verbose, raise_error=raise_error)
        # print("The catalogue for the requested data has not been acquired.")


def get_category_menu(name, update=False, confirmation_required=True, verbose=False,
                      raise_error=False):
    """
    Gets a menu of the available classes from the specified URL.

    This function scrapes a web page for available classes (typically categorised hyperlinks) and
    returns them as a dictionary. It also provides options to update the catalogue and
    save it as a JSON file.

    :param name: The name of the data category.
    :type name: str
    :param update: Whether to check for updates to the package data; defaults to ``True``.
    :type update: bool
    :param confirmation_required: Whether user confirmation is required before proceeding;
        defaults to ``True``.
    :type confirmation_required: bool
    :param verbose: Whether to print relevant information to the console; defaults to ``False``.
    :type verbose: bool | int
    :param raise_error: Whether to raise the provided exception;
        if ``raise_error=False`` (default), the error will be suppressed.
    :type raise_error: bool
    :return: A category menu in dictionary form,
        where keys are data cluster names and values are URLs.
    :rtype: dict | None

    **Examples**::

        >>> from pyrcs.parser import get_category_menu
        >>> menu = get_category_menu(name='Line data')
        >>> type(menu)
        dict
        >>> list(menu.keys())
        ['Line data']
        >>> len(menu['Line data'])
        7
    """

    path_to_file = cd_data("catalogue", f"{name.lower().replace(' ', '-')}-menu.json", mkdir=True)

    if os.path.isfile(path_to_file) and not update:
        return load_data(path_to_file)

    if confirmed("To collect/update category menu?", confirmation_required=confirmation_required):
        try:
            source = requests.get(url=homepage_url(), headers=fake_requests_headers())
            source.raise_for_status()
        except Exception as e:
            print_instance_connection_error(
                update=update, verbose=True if update else verbose, e=e, raise_error=raise_error)
            return None

        try:
            soup = bs4.BeautifulSoup(markup=source.content, features='html.parser')

            drop_btn_ = soup.select(f'button:-soup-contains("{name}")')
            drop_btn = drop_btn_[0]

            a_href_list = drop_btn.find_next_sibling('div').find_all('a')

            cls_menu_ = [
                (a.get_text(), urllib.parse.urljoin(homepage_url(), a['href']))
                for a in a_href_list]

            cls_menu = {name: dict(cls_menu_)}

            save_data(cls_menu, path_to_file, indent=4, verbose=(verbose == 2 or False))

            return cls_menu

        except Exception as e:
            _print_failure_message(
                e, prefix="Failed. Error:", verbose=verbose, raise_error=raise_error)

    else:
        if verbose in {True, 1}:
            print("Cancelled.")


def get_heading_text(heading_tag, elem_tag_name='em'):
    """
    Gets the text from a given HTML heading tag.

    :param heading_tag: The HTML tag of a heading element.
    :type heading_tag: bs4.element.Tag
    :param elem_tag_name: The tag name of an inner element within the heading; defaults to ``'em'``.
    :type elem_tag_name: str
    :return: Cleaned text of the heading tag.
    :rtype: str

    **Examples**::

        >>> from pyrcs.parser import get_heading_text
        >>> from pyrcs.line_data import Electrification
        >>> elec = Electrification()
        >>> url = elec.catalogue[elec.KEY_TO_INDEPENDENT_LINES]
        >>> source = requests.get(url=url, headers=fake_requests_headers())
        >>> soup = bs4.BeautifulSoup(markup=source.content, features='html.parser')
        >>> h3 = soup.find('h3')
        >>> h3_text = get_heading_text(heading_tag=h3, elem_tag_name='em')
        >>> h3_text
        'Beamish Tramway'
    """

    heading_x = []

    for elem in heading_tag.contents:
        # noinspection PyUnresolvedReferences
        if elem.name == elem_tag_name:
            heading_x.append('[' + elem.text + ']')
        else:
            heading_x.append(elem.text)

    heading_text = ''.join(heading_x)

    return heading_text


def get_page_catalogue(url, head_tag_name='nav', head_tag_txt='Jump to:', feature_tag_name='h3',
                       verbose=False, raise_error=False):
    """
    Gets the catalogue of features from the main page of a data cluster.

    This function extracts structured data (features) from a web page by parsing specific tags,
    typically used for features like headings and links in railway-related databases.

    :param url: The URL of the main page of a data cluster.
    :type url: str
    :param head_tag_name: The tag name of the feature list at the top of the page;
        defaults to ``'nav'``.
    :type head_tag_name: str
    :param head_tag_txt: Text contained in the head tag; defaults to ``'Jump to: '``.
    :type head_tag_txt: str
    :param feature_tag_name: The tag name of the headings of each feature; defaults to ``'h3'``.
    :type feature_tag_name: str
    :param verbose: Whether to print relevant information to the console; defaults to ``False``.
    :type verbose: bool | int
    :param raise_error: Whether to raise the provided exception;
        if ``raise_error=False`` (default), the error will be suppressed.
    :type raise_error: bool
    :return: A dataframe containing the page's feature catalogue with columns for feature, URL and
        heading.
    :rtype: pandas.DataFrame

    **Examples**::

        >>> from pyrcs.parser import get_page_catalogue
        >>> from pyhelpers.settings import pd_preferences
        >>> pd_preferences(max_columns=1)
        >>> elec_url = 'http://www.railwaycodes.org.uk/electrification/mast_prefix2.shtm'
        >>> elec_catalogue = get_page_catalogue(elec_url)
        >>> elec_catalogue
                                                      Feature  ...
        0                                     Beamish Tramway  ...
        1                                  Birkenhead Tramway  ...
        2                         Black Country Living Museum  ...
        3                                   Blackpool Tramway  ...
        4   Brighton and Rottingdean Seashore Electric Rai...  ...
        ..                                                ...  ...
        17                                     Seaton Tramway  ...
        18                                Sheffield Supertram  ...
        19                          Snaefell Mountain Railway  ...
        20  Summerlee, Museum of Scottish Industrial Life ...  ...
        21                                  Tyne & Wear Metro  ...
        [22 rows x 3 columns]
        >>> elec_catalogue.columns.to_list()
        ['Feature', 'URL', 'Heading']
    """

    try:
        source = requests.get(url=url, headers=fake_requests_headers())
        source.raise_for_status()
    except Exception as e:
        print_instance_connection_error(verbose=verbose, e=e, raise_error=raise_error)
        return None

    try:
        soup = bs4.BeautifulSoup(markup=source.content, features='html.parser')

        page_catalogue = pd.DataFrame({'Feature': [], 'URL': [], 'Heading': []})

        for nav in soup.find_all(head_tag_name):
            nav_text = nav.text.replace('\r\n', '').strip()

            if re.match(r'^({})'.format(head_tag_txt), nav_text):
                sep = '\xa0| ' if '\xa0| ' in nav_text else '\n'
                feature_names = nav_text.replace(f'{head_tag_txt}{sep}', '').split(sep)
                page_catalogue['Feature'] = feature_names

                feature_urls = []
                for item_name in feature_names:
                    text_pat = re.compile(r'.*{}.*'.format(item_name), re.IGNORECASE)
                    a = nav.find('a', string=text_pat)

                    feature_urls.append(urllib.parse.urljoin(url, a.get('href')))

                page_catalogue['URL'] = feature_urls

        feature_headings = []
        for h3 in soup.find_all(feature_tag_name):
            sub_heading = get_heading_text(heading_tag=h3, elem_tag_name='em')
            feature_headings.append(sub_heading)

        page_catalogue['Heading'] = feature_headings

        return page_catalogue

    except Exception as e:
        _print_failure_message(e, verbose=verbose, raise_error=raise_error)


def get_hypertext(hypertext_tag, hyperlink_tag_name='a', md_style=True):
    """
    Gets hyperlinked text from a specified HTML tag.

    This function scrapes hypertext content, optionally returning it in Markdown format if
    requested.

    :param hypertext_tag: The tag containing hyperlinked text.
    :type hypertext_tag: bs4.element.Tag | bs4.element.PageElement
    :param hyperlink_tag_name: The tag name of the hyperlink within the hypertext;
        defaults to ``'a'``.
    :type hyperlink_tag_name: str
    :param md_style: Whether to return the hypertext in Markdown style, defaults to ``True``.
    :type md_style: bool
    :return: The hypertext.
    :rtype: str

    **Examples**::

        >>> from pyrcs.parser import get_hypertext
        >>> from pyrcs.line_data import Electrification
        >>> import bs4
        >>> import requests
        >>> elec = Electrification()
        >>> url = elec.catalogue[elec.KEY_TO_INDEPENDENT_LINES]
        >>> source = requests.get(url)
        >>> soup = bs4.BeautifulSoup(source.content, 'html.parser')
        >>> h3 = soup.find('h3')
        >>> p = h3.find_all_next('p')[8]
        >>> p
        <p>Croydon Tramlink mast references can be found on the <a href="http://www.croydon-traml...
        >>> hyper_txt = get_hypertext(hypertext_tag=p, md_style=True)
        >>> hyper_txt
        'Croydon Tramlink mast references can be found on the [Croydon Tramlink Unofficial Site](...
    """

    hypertext_x = []

    for x in hypertext_tag.contents:
        # noinspection PyUnresolvedReferences
        if x.name == hyperlink_tag_name:
            # noinspection PyUnresolvedReferences
            href = x.get('href')
            x_text = '[' + x.text + ']' + f'({href})' if md_style else x.text + f' ({href})'
            hypertext_x.append(x_text)
        else:
            hypertext_x.append(x.text)

    hypertext_tag = ''.join(hypertext_x).replace('\xa0', '').replace('  ', ' ')

    return hypertext_tag



"""
Collects data of British `railway bridges <http://www.railwaycodes.org.uk/bridges/bridges0.shtm>`_.
"""

import os
import re
import urllib.parse

import bs4
from pyhelpers.ops import split_list_by_size

from .._base import _Base
from ..parser import _get_last_updated_date
from ..utils import homepage_url


class Bridges(_Base):
    """
    A class for collecting data of
    `railway bridges <http://www.railwaycodes.org.uk/bridges/bridges0.shtm>`_.
    """

    #: The name of the data.
    NAME: str = 'Railway bridges'
    #: The key for accessing the data.
    KEY: str = 'Bridges'
    #: The URL of the main webpage for the data.
    URL: str = urllib.parse.urljoin(homepage_url(), '/bridges/bridges0.shtm')
    #: The key used to reference the last updated date in the data.
    KEY_TO_LAST_UPDATED_DATE: str = 'Last updated date'

    def __init__(self, data_dir=None, update=False, verbose=True):
        """
        :param data_dir: Directory where the data is stored; defaults to ``None``.
        :type data_dir: str | None
        :param verbose: Whether to print relevant information to the console; defaults to ``True``.
        :type verbose: bool | int

        :ivar dict catalogue: The catalogue of the data.
        :ivar str last_updated_date: The date when the data was last updated.
        :ivar str data_dir: The path to the directory containing the data.
        :ivar str current_data_dir: The path to the current data directory.

        **Examples**::

            >>> from pyrcs.line_data import Bridges  # from pyrcs import Bridges
            >>> bdg = Bridges()
            >>> bdg.NAME
            'Railway bridges'
            >>> bdg.URL
            'http://www.railwaycodes.org.uk/bridges/bridges0.shtm'
        """

        super().__init__(
            data_dir=data_dir, content_type='introduction', data_category="line-data",
            update=update, verbose=verbose)

    def _parse_h4_ul_li(self, h4_ul_li):
        h4_ul_li_contents = h4_ul_li.contents

        h4_ul_li_dict = {}
        if len(h4_ul_li_contents) == 1:
            h4_ul_li_content = h4_ul_li_contents[0]

            text = h4_ul_li_content.get_text(strip=True)
            href = h4_ul_li_content.get('href')

        else:  # len(h4_ul_li_contents) == 2:
            span_a_href, suppl_text = h4_ul_li_contents
            if not isinstance(suppl_text, str):
                suppl_text, span_a_href = h4_ul_li_contents

            text = span_a_href.get_text(strip=True)
            if suppl_text:
                text += ' ' + suppl_text

            href = span_a_href.find('a').get('href')

        link = urllib.parse.urljoin(os.path.dirname(self.URL) + '/', href)
        h4_ul_li_dict.update({text: link})

        return h4_ul_li_dict

    def _parse_h4(self, h4):
        h4_txt = h4.get_text(strip=True)

        h4_dat = None

        h4_ul = h4.find_next(name='ul')
        if isinstance(h4_ul, bs4.Tag):
            h4_ul_lis = h4_ul.find_all(name='li')
            h4_dat = {}
            for h4_ul_li in h4_ul_lis:
                h4_dat.update(self._parse_h4_ul_li(h4_ul_li))

        elif h4_ul is None:
            h4_pre = h4.find_next('pre')
            if isinstance(h4_pre, bs4.Tag):
                h4_dat = dict([x.split('\t') for x in h4_pre.text.split('\n')])

        return {h4_txt: h4_dat}

    def _parse_source(self, source, verbose=False):
        """
        Scrapes codes of `railway bridges`_ from its source webpage.

        .. _`railway bridges`: http://www.railwaycodes.org.uk/bridges/bridges0.shtm

        :param source: HTTP response containing the webpage content.
        :type source: requests.Response
        :param verbose: Whether to print relevant information in the console; defaults to ``False``.
        :type verbose: bool | int
        :return: A dictionary containing railway bridge data and the date of the last update.
        :rtype: dict
        """

        soup = bs4.BeautifulSoup(markup=source.content, features='html.parser')

        h4_list = soup.find_all(name='h4')

        data = {k: v for h4 in h4_list for k, v in self._parse_h4(h4).items()}

        # Key to text presentation conventions
        keys_h3 = h4_list[-1].find_next(name='h3')
        keys_p_contents = keys_h3.find_next('p').contents

        keys_p_contents_ = []
        for x in keys_p_contents:
            if isinstance(x, str):
                y = re.sub(r'( = +)|\n', '', x).capitalize()
            else:
                y = x.get_text(strip=True)
            keys_p_contents_.append(y)

        sub_dict = split_list_by_size(keys_p_contents_, sub_len=2)
        keys_dict = {keys_h3.text: {k: v for k, v in sub_dict}}

        data.update(keys_dict)

        last_updated_date = _get_last_updated_date(soup=soup, parsed=True)
        data = {self.KEY: data, self.KEY_TO_LAST_UPDATED_DATE: last_updated_date}

        if verbose in {True, 1}:
            print("Done.")

        self._save_data_to_file(data, data_name=self.KEY, ext=".json", verbose=verbose, indent=4)

        return data

    def collect_codes(self, confirmation_required=True, verbose=False, raise_error=False):
        """
        Collects codes of `railway bridges`_ from its source webpage.

        .. _`railway bridges`: http://www.railwaycodes.org.uk/bridges/bridges0.shtm

        :param confirmation_required: Whether user confirmation is required;
            if ``confirmation_required=True`` (default), prompts the user for confirmation
            before proceeding with data collection.
        :type confirmation_required: bool
        :param verbose: Whether to print relevant information in the console; defaults to ``False``.
        :type verbose: bool | int
        :param raise_error: Whether to raise the provided exception;
            if ``raise_error=False`` (default), the error will be suppressed.
        :type raise_error: bool
        :return: A dictionary containing railway bridge data and the date of the last update,
            or ``None`` if no data is collected.
        :rtype: dict | None

        **Examples**::

            >>> from pyrcs.line_data import Bridges  # from pyrcs import Bridges
            >>> bdg = Bridges()
            >>> bdg_codes = bdg.collect_codes(verbose=True)
            To collect data of railway bridges
            ? [No]|Yes: yes
            >>> type(bdg_codes)
            dict
            >>> list(bdg_codes.keys())
            ['East Coast Main Line',
             'Midland Main Line',
             'West Coast Main Line',
             'Scotland',
             'Elizabeth Line',
             'London Overground',
             'Anglia',
             'London Underground',
             'Key to text presentation conventions']
            >>> bdg_codes['Key to text presentation conventions']
            {'Bold': 'Existing bridges',
             'Bold italic': 'Existing locations',
             'Light italic': 'Former/historical locations',
             'Red': 'Stations',
             'Deep red': 'Level crossings',
             'Brown': 'Ventilation shafts',
             'Purple': 'Junctions',
             'Black,grey': 'Bridges and culverts',
             'Green': 'Tunnel portals',
             'Bright blue': 'Viaducts',
             'Deep blue': 'Boundaries'}
        """

        data = self._collect_data_from_source(
            data_name=self.NAME.lower(), method=self._parse_source, url=self.URL,
            confirmation_required=confirmation_required, verbose=verbose, raise_error=raise_error)

        return data

    def fetch_codes(self, update=False, dump_dir=None, verbose=False, **kwargs):
        """
        Fetches codes of `railway bridges`_.

        .. _`railway bridges`: http://www.railwaycodes.org.uk/bridges/bridges0.shtm

        :param update: Whether to check for updates to the package data; defaults to ``False``.
        :type update: bool
        :param dump_dir: The path to a directory where the data file will be saved;
            defaults to ``None``.
        :type dump_dir: str | None
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :return: A dictionary containing railway bridge data and the date of the last update,
            or ``None`` if no data is retrieved.
        :rtype: dict | None

        **Examples**::

            >>> from pyrcs.line_data import Bridges  # from pyrcs import Bridges
            >>> bdg = Bridges()
            >>> bdg_codes = bdg.fetch_codes()
            >>> type(bdg_codes)
            dict
            >>> list(bdg_codes.keys())
            ['East Coast Main Line',
             'West Coast Main Line',
             'Scotland',
             'Elizabeth Line',
             'London Overground',
             'Anglia',
             'London Underground',
             'Addendum',
             'Key to text presentation conventions']
            >>> bdg_codes['Key to text presentation conventions']
            {'Bold': 'Existing bridges',
             'Bold italic': 'Existing locations',
             'Light italic': 'Former/historical locations',
             'Red': 'Stations',
             'Deep red': 'Level crossings',
             'Brown': 'Ventilation shafts',
             'Purple': 'Junctions',
             'Black,grey': 'Bridges and culverts',
             'Green': 'Tunnel portals',
             'Bright blue': 'Viaducts',
             'Deep blue': 'Boundaries'}
        """

        args = {
            'data_name': self.KEY,
            'method': self.collect_codes,
            'ext': ".json",
        }
        kwargs.update(args)

        bridges_data = self._fetch_data_from_file(
            update=update, dump_dir=dump_dir, verbose=verbose, **kwargs)

        return bridges_data



"""
Collects codes of infrastructure features.

This category includes:

    - `HABD and WILD <http://www.railwaycodes.org.uk/misc/habdwild.shtm>`_
    - `Water troughs <http://www.railwaycodes.org.uk/misc/troughs.shtm>`_
    - `Telegraph codes <http://www.railwaycodes.org.uk/misc/telegraph.shtm>`_
    - `Driver/guard buzzer codes <http://www.railwaycodes.org.uk/misc/buzzer.shtm>`_
"""

import inspect
import itertools

from .buzzer import Buzzer
from .habd_wild import HabdWild
from .telegraph import Telegraph
from .trough import WaterTroughs
from .._base import _Base
from ..utils import get_batch_fetch_verbosity, is_homepage_connectable


class Features(_Base):
    """
    A class for collecting data of several infrastructure features, including
    *HABDs and WILDs*, *water troughs locations*, *telegraph code words* and *buzzer codes*.
    """

    #: The name of the data.
    NAME: str = 'Infrastructure features'
    #: The key for accessing the data.
    KEY: str = 'Features'

    #: The key for accessing the data of *HABD* and *WILD*.
    KEY_TO_HABD_WILD: str = HabdWild.KEY
    #: The key for accessing the data of *water troughs*.
    KEY_TO_TROUGH: str = WaterTroughs.KEY
    #: The key for accessing the data of *telegraph codes*.
    KEY_TO_TELEGRAPH: str = Telegraph.KEY
    #: The key for accessing the data of *buzzer codes*.
    KEY_TO_BUZZER: str = Buzzer.KEY

    #: The URL of the main web page for the data.
    URL: str = HabdWild.URL

    #: The key used to reference the last updated date in the data.
    KEY_TO_LAST_UPDATED_DATE: str = 'Last updated date'

    def __init__(self, data_dir=None, update=False, verbose=True):
        """
        :param data_dir: The name of the directory for storing the data; defaults to ``None``.
        :type data_dir: str | None
        :param update: Whether to check for updates to the catalogue; defaults to ``False``.
        :type update: bool
        :param verbose: Whether to print relevant information to the console; defaults to ``True``.
        :type verbose: bool | int

        :ivar dict catalogue: The catalogue of the data.
        :ivar str last_updated_date: The date when the data was last updated.
        :ivar str data_dir: The path to the directory containing the data.
        :ivar str current_data_dir: The path to the current data directory.

        **Examples**::

            >>> from pyrcs.other_assets import Features  # from pyrcs import Features
            >>> feats = Features()
            >>> feats.NAME
            'Infrastructure features'
        """

        super().__init__(
            data_dir=data_dir, content_type='catalogue', data_category="other-assets",
            update=update, verbose=verbose)

    @staticmethod
    def _get_classes_in_module(verbose=False):
        # noinspection PyShadowingNames
        """
        Retrieve all classes imported within the current module, excluding locally defined classes.

        This function inspects the global namespace of the module to find all imported classes,
        excluding those defined within the module itself. The ``_Base`` class is also excluded.

        :param verbose: Whether to print the imported class names and objects to the console.
        :type verbose: bool
        :return: A list of tuples containing the names and objects of imported classes.
        :rtype: list[tuple[str, type]]

        **Examples**::

            >>> from pyrcs.other_assets import Features  # from pyrcs import Features
            >>> feats = Features()
            >>> imported_classes = feats._get_classes_in_module()
            >>> imported_classes
            [('HABDWILD', pyrcs.other_assets.habd_wild.HABDWILD),
             ('Buzzer', pyrcs.other_assets.buzzer.Buzzer),
             ('Telegraph', pyrcs.other_assets.telegraph.Telegraph),
             ('WaterTroughs', pyrcs.other_assets.trough.WaterTroughs)]
        """

        # Get the current module's global namespace
        current_module = globals()

        # Filter for classes
        imported_classes = []
        for name, obj in current_module.items():
            if inspect.isclass(obj) and obj.__module__ != __name__ and name != '_Base':
                imported_classes.append((name, obj))

        if verbose:  # Get and print all classes
            for name, cls in imported_classes:
                print(f"Imported class name: {name}, Imported class: {cls}")

        return imported_classes

    def fetch_codes(self, update=False, dump_dir=None, verbose=False, **kwargs):
        """
        Fetches data of infrastructure features.

        Including:

            - `HABD and WILD <http://www.railwaycodes.org.uk/misc/habdwild.shtm>`_
            - `Water troughs <http://www.railwaycodes.org.uk/misc/troughs.shtm>`_
            - `Telegraph codes <http://www.railwaycodes.org.uk/misc/telegraph.shtm>`_
            - `Driver/guard buzzer codes <http://www.railwaycodes.org.uk/misc/buzzer.shtm>`_

        :param update: Whether to check for updates to the package data; defaults to ``False``.
        :type update: bool
        :param dump_dir: The path to a directory where the data file will be saved;
            defaults to ``None``.
        :type dump_dir: str | None
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :return: A dictionary containing the data of features codes and
            the date they were last updated.
        :rtype: dict

        **Examples**::

            >>> from pyrcs.other_assets import Features  # from pyrcs import Features
            >>> feats = Features()
            >>> feats_codes = feats.fetch_codes()
            >>> type(feats_codes)
            dict
            >>> list(feats_codes.keys())
            ['Features', 'Last updated date']
            >>> feats.KEY
            'Features'
            >>> feats_codes_dat = feats_codes[feats.KEY]
            >>> type(feats_codes_dat)
            dict
            >>> list(feats_codes_dat.keys())
            ['Buzzer codes', 'HABD and WILD', 'Telegraphic codes', 'Water troughs']
            >>> water_troughs_locations = feats_codes_dat[feats.KEY_TO_TROUGH]
            >>> type(water_troughs_locations)
            pandas.core.frame.DataFrame
            >>> water_troughs_locations.head()
                ELR  ... Length (Yard)
            0   BEI  ...           NaN
            1   BHL  ...    620.000000
            2  CGJ2  ...      0.666667
            3  CGJ6  ...    561.000000
            4  CGJ6  ...    560.000000
            [5 rows x 6 columns]
            >>> hw_codes_dat = feats_codes_dat[feats.KEY_TO_HABD_WILD]
            >>> type(hw_codes_dat)
            dict
            >>> list(hw_codes_dat.keys())
            ['HABD', 'WILD']
            >>> habd_dat = hw_codes_dat['HABD']
            >>> type(habd_dat)
            pandas.core.frame.DataFrame
            >>> habd_dat.head()
                ELR  ...                                              Notes
            0  BAG2  ...
            1  BAG2  ...  installed 29 September 1997, later moved to 74...
            2  BAG2  ...                             previously at 74m 51ch
            3  BAG2  ...                          removed 29 September 1997
            4  BAG2  ...           present in 1969, later moved to 89m 00ch
            [5 rows x 5 columns]
            >>> wild_dat = hw_codes_dat['WILD']
            >>> type(wild_dat)
            pandas.core.frame.DataFrame
            >>> wild_dat.head()
                ELR  ...                                              Notes
            0  AYR3  ...
            1  BAG2  ...
            2  BML1  ...
            3  BML1  ...
            4  CGJ3  ...  moved to 183m 68ch from 8 September 2018 / mov...
            [5 rows x 5 columns]
        """

        verbose_ = get_batch_fetch_verbosity(data_dir=dump_dir, verbose=verbose)

        features_codes_dat = []

        for (_, cls) in self._get_classes_in_module():
            dat = cls().fetch_codes(
                update=update, verbose=verbose_ if is_homepage_connectable() else False, **kwargs)
            features_codes_dat.append(dat)

        features_codes = {
            self.KEY: {next(iter(x)): next(iter(x.values())) for x in features_codes_dat},
            self.KEY_TO_LAST_UPDATED_DATE:
                max(next(itertools.islice(iter(x.values()), 1, 2)) for x in features_codes_dat),
        }

        if dump_dir is not None:
            self._save_data_to_file(
                data=features_codes, data_name=self.KEY, dump_dir=dump_dir, verbose=verbose)

        return features_codes



"""
Collects data of
`CRS, NLC, TIPLOC and STANOX codes <http://www.railwaycodes.org.uk/crs/crs0.shtm>`_.
"""

import collections
import re
import string
import urllib.parse

import bs4
import pandas as pd
import requests
from pyhelpers._cache import _print_failure_message
from pyhelpers.dirs import validate_dir
from pyhelpers.ops import fake_requests_headers

from .._base import _Base
from ..parser import _get_last_updated_date, get_page_catalogue, parse_tr
from ..utils import format_confirmation_prompt, get_collect_verbosity_for_fetch, homepage_url, \
    is_homepage_connectable, print_instance_connection_error, print_void_collection_message, \
    validate_initial


def _parse_raw_location_name(x):
    """
    Parses the location name and extracts any associated note from the raw data.

    This function separates the main location name from metadata/notes usually
    found in parentheses, brackets or separated by specific markers.

    :param x: Location name (in raw data).
    :type x: str | None
    :return: A tuple of (Location Name, Note). Returns ``('', '')`` if input is None/Empty.
    :rtype: tuple

    **Examples**::

        >>> from pyrcs.line_data.loc_id import _parse_raw_location_name
        >>> _parse_raw_location_name(None)
        ('', '')
        >>> _parse_raw_location_name('Abbey Wood')
        ('Abbey Wood', '')
        >>> _parse_raw_location_name('Abercynon (formerly Abercynon South)')
        ('Abercynon', 'formerly Abercynon South')
        >>> _parse_raw_location_name('Allerton (reopened as Liverpool South Parkway)')
        ('Allerton', 'reopened as Liverpool South Parkway')
        >>> _parse_raw_location_name('Ashford International [domestic portion]')
        ('Ashford International', 'domestic portion')
        >>> _parse_raw_location_name('Ayr [unknown feature]')
        ('Ayr', 'unknown feature')
        >>> _parse_raw_location_name('Birkenhead Hamilton Square [see Hamilton Square]')
        ('Birkenhead Hamilton Square', 'see Hamilton Square')
    """

    if not x:
        return '', ''

    x_ = x.strip()  # Clean leading/trailing whitespace

    # Handle explicit special delimiters
    if '✖' in x_:  # Check for the '✖' symbol
        name, _, note = x_.partition('✖')
        return name.strip(), note.strip()

    if 'STANOX ' in x_:  # Check for 'STANOX'
        # Split at the first occurrence of STANOX
        name, sep, note = x_.partition('STANOX')
        return name.strip(), (sep + note).strip()

    # Regex to capture: Name (Note Content) OR Name [Note Content]
    match = re.search(r'^(.*?)\s*[(\[](.+)[)\]]$', x_, re.DOTALL)

    if match:
        name_part, note_part = match.groups()

        # Keywords that identify a bracket/parenthesis as a 'Note'
        note_keywords = [
            'originally', 'formerly', 'later', 'presumed', 'was', 'reopened',
            'portion', 'see', 'unknown feature', 'definition unknown', 'now deleted', 'now'
        ]
        keyword_pattern = re.compile('|'.join(note_keywords), re.IGNORECASE)

        # Split if it contains a keyword OR if it's the standard format from your docs
        if keyword_pattern.search(note_part):
            # Use regex to strip any leading/trailing non-alphanumeric noise like ["( or )"]
            note_part = re.sub(r'^[^a-zA-Z0-9]+|[^a-zA-Z0-9]+$', '', note_part)
            return name_part.strip(), note_part.strip()

    return x_, ''


def _amendment_to_location_names():
    """
    Creates a dictionary of amendments to adjust location names using regular expressions.

    This method generates a dictionary where the keys are regular expression patterns, and the
    values are the corresponding amendments to be applied to location names.

    :return: A dictionary where each key is a regular expression pattern and each value is the
        corresponding replacement string to amend the location names.
    :rtype: dict

    **Examples**::

        >>> from pyrcs.line_data.loc_id import _amendment_to_location_names
        >>> loc_name_amendment_dict = _amendment_to_location_names()
        >>> list(loc_name_amendment_dict.keys())
        ['Location']
    """

    location_name_amendment_dict = {
        'Location': {re.compile(r' And | \+ '): ' & ',
                     re.compile(r'-By-'): '-by-',
                     re.compile(r'-In-'): '-in-',
                     re.compile(r'-En-Le-'): '-en-le-',
                     re.compile(r'-La-'): '-la-',
                     re.compile(r'-Le-'): '-le-',
                     re.compile(r'-On-'): '-on-',
                     re.compile(r'-The-'): '-the-',
                     re.compile(r' Of '): ' of ',
                     re.compile(r'-Super-'): '-super-',
                     re.compile(r'-Upon-'): '-upon-',
                     re.compile(r'-Under-'): '-under-',
                     re.compile(r'-Y-'): '-y-'}}

    return location_name_amendment_dict


def _extra_annotations():
    extra_annotations = [
        ('✖Earlier code ', '✖Later code'),
        ('✖Later code ', '✖Earlier code'),
        ('✖Earlier code, soon replaced ', '✖Later code introduced before station opened'),
        ('✖Later code introduced before station opened ', '✖Earlier code, soon replaced'),
        ('✖Original code ', '✖Later code'),
        ('✖Later code ', '✖Original code'),
        ('✖Original code ', '✖Later code from station opening'),
        ('✖Later code from station opening ', '✖Original code'),
        ('✖Newer designation? ', '✖Older designation?'),
        ('✖Older designation? ', '✖Newer designation?'),
        ('✖Both codes quoted with equal reliability ',
         '✖Both codes quoted with equal reliability'),
        ('✖Code used by operational research software ',
         '✖Code as used by National Rail Enquiries and other public systems'),
        ('✖Code as used by National Rail Enquiries and other public systems ',
         '✖Code used by operational research software'),
        ('✖Code assigned in error ', '✖Corrected code'),
        ('✖Original code ', '✖Later code after becoming part of national network'),
        ('✖Later code after becoming part of national network ', '✖Original code'),
        ('✖Code used after station reopened ', '✖Code used until station reopened'),
        ('✖Code used until station reopened ', '✖Code used after station reopened'),
        ('✖Code used after nearby station reopened ', '✖Code used until nearby station reopened'),
        ('✖Code used until nearby station reopened ', '✖Code used after nearby station reopened'),
        ('✖Original code; see also CRS explanation ', '✖Later code; see also CRS explanation'),
        ('✖Later code; see also CRS explanation ', '✖Original code; see also CRS explanation'),
        ('✖Most sources state LOUNOC, some have LOUNGOC ',
         '✖Most sources state LOUNOC, some have LOUNGOC'),
        ("✖Code may start with 'R' ", "✖Code may start with 'B'"),
        ("✖Code may start with 'B' ", "✖Code may start with 'R'"),
        ('✖Older designation when station open ', '✖Newer designation once station closed'),
        ('✖Newer designation once station closed ', '✖Older designation when station open'),
        ('✖Original code when named Butlins Penychain ', '✖Later code after renaming Penychain'),
        ('✖Later code after renaming Penychain ', '✖Original code when named Butlins Penychain'),
        ('✖Earlier code ',
         '✖Later code after East London Line operated as part of national network'),
        ('✖Code should be ROODENDGD but some listings show this as RO0DENDGD (with zero digit) ',
         '✖Code should be ROODENDGD but some listings show this as RO0DENDGD (with zero digit)'),
        ('✖Earlier code? ', '✖Later code'),
        ('✖See CRS explanation ', '✖Code believed listed in error'),
        ('✖Code is 56582 but some sources have 56782 ',
         '✖Code is 56582 but some sources have 56782'),
        ('✖Code should be 79251 but one source has 79521 ',
         '✖Code should be 79251 but one source has 79521'),
        ('✖Code is 56540 but some sources have 56450 ',
         '✖Code is 56540 but some sources have 56450'),
        ('✖Earlier code ', '✖Later code; also see CRS explanation'),
        ('✖Later code ', '✖Possibly earlier code'),
        ('✖Code is 142505 but is also reported as 145505 ',
         '✖Code is 142505 but is also reported as 145505'),
        ('✖Earlier code ', '✖Later code after station closed'),
    ]

    return extra_annotations


def _count_sep(x):
    if '\r\n' in x:
        r_n_counts = x.count('\r\n')
    elif '\r' in x:
        r_n_counts = x.count('\r')
    else:  # Ad hoc
        if '~LO\n' in x:
            x = x.replace('~LO\n', '')
        r_n_counts = x.count('\n')
    return r_n_counts


def _split_dat_and_note(x):
    if '\r\n' in x:
        x_ = x.split('\r\n')

    elif '\r' in x:
        x_ = x.split('\r')

    elif '\n' in x:
        if '~LO\n' in x:  # Ad hoc
            x = x.replace('~LO\n', '')
        x_ = x.split('\n')

    else:
        x_ = x

    return x_


def _fix_exceptional_cases(data):
    x1 = 'Ely Papworth Sidings English Welsh & Scottish Railway International\n' \
         'Ely Papworth Sidings DB Schenker International'

    if x1 in data['Location'].values:
        x1_ = 'Ely Papworth Sidings English Welsh & Scottish Railway International\n' \
              'Ely Papworth Sidings DB Schenker International\n' \
              'Ely Papworth Sidings DB Schenker International'

        data.loc[data['Location'] == x1, 'Location'] = x1_

    return data


def _parse_mult_alt_codes(data):
    """
    Cleanses multiple alternatives for every code column.

    :param data: The preprocessed data of the location codes.
    :type data: pandas.DataFrame
    :return: The cleansed data of the location codes where multiple alternatives are replicated.
    :rtype: pandas.DataFrame
    """

    df = data.copy()
    df = _fix_exceptional_cases(df)

    code_cols = ['Location', 'CRS', 'NLC', 'TIPLOC', 'STANME', 'STANOX']

    r_n_counts = df[code_cols].map(_count_sep)
    # # Debugging:
    # for col in code_col_names:
    #     for i, x in enumerate(data_[col]):
    #         try:
    #             lid._count_sep(x)
    #         except Exception:
    #             print(col, i, x)
    #             break
    r_n_counts_ = r_n_counts.mul(-1).add(r_n_counts.max(axis=1), axis='index')

    for col in code_cols:
        for i in df.index:
            d = r_n_counts_.loc[i, col]
            val = df.loc[i, col]
            if d > 0:
                if '\r\n' in val:
                    if col == 'Location':
                        df.loc[i, col] = val + ''.join(['\r\n' + val.split('\r\n')[-1]] * d)
                    else:
                        df.loc[i, col] = val + ''.join(['\r\n'] * d)
                elif '\r' in val:
                    if col == 'Location':
                        df.loc[i, col] = val + ''.join(['\r' + val.split('\r')[-1]] * d)
                    else:
                        df.loc[i, col] = val + ''.join(['\r'] * d)
                else:  # e.g. '\n' in dat:
                    if col == 'Location':
                        df.loc[i, col] = '\n'.join([val] * (d + 1))
                    else:
                        df.loc[i, col] = val + ''.join(['\n'] * d)

    df[code_cols] = df[code_cols].map(_split_dat_and_note)

    df = df.explode(code_cols, ignore_index=True)

    temp = df.select_dtypes(['object'])
    df[temp.columns] = temp.apply(lambda x_: x_.str.strip())

    return df


def _parse_code_note(x):
    """
    Gets note for every code column.

    :param x: The raw data of a given code.
    :type x: str | None
    :return: The extra information (if any) about the given code.
    :rtype: str

    **Examples**::

        >>> from pyrcs.line_data.loc_id import _parse_code_note
        >>> _parse_code_note('860260✖Earlier code')
        ('860260', 'Earlier code')
    """

    if not x:
        y, note = x, ''

    else:
        if '✖' in x:
            y, note = map(str.strip, x.split('✖', 1))

        else:  # Search for notes
            n1 = re.search(r'(?<=[\[(])[\w,? ]+(?=[)\]])', x)

            if n1:
                note = n1.group(0)
                y = x.replace(note, '').strip('[(\')] ')

                n2 = re.search(r'[\w ,]+(?= [\[(\'])', note)  # Remove redundant characters
                if n2:
                    note = n2.group(0)

            else:
                y, note = x.strip(), ''  # Default case: No note found

    return y, note


def _parse_code_notes(data):
    """
    Gets notes for every code column.

    :param data: The preprocessed data of the location codes.
    :type data: pandas.DataFrame
    """

    codes_col_names = ['CRS', 'NLC', 'TIPLOC', 'STANME', 'STANOX']

    for col in codes_col_names:
        data[[col, col + '_Note']] = pd.DataFrame(
            data[col].map(_parse_code_note).to_list(), index=data.index)
    # # Debugging:
    # for col in codes_col_names:
    #     for i, x in enumerate(data[col]):
    #         try:
    #             _parse_code_note(x)
    #         except Exception:
    #             print(col, i, x)
    #             break
    return data


def _stanox_note(x):
    """
    Parses STANOX note.

    :param x: STANOX note.
    :type x: str | None
    :return: STANOX and its corresponding note.
    :rtype: tuple
    """

    if x in ('-', '') or x is None:
        stanox, note = '', ''

    else:
        if re.match(r'\d{5}$', x):
            stanox, note = x, ''

        elif re.match(r'\d{5}\*$', x):
            stanox, note = x.rstrip('*'), 'Pseudo STANOX'

        elif re.match(r'\d{5} \w.*', x):
            stanox = re.search(r'\d{5}', x).group()
            note = re.search(r'(?<= )\w.*', x).group()

        else:
            d = re.search(r'[\w *,]+(?= [\[(\'])', x)
            stanox = d.group().strip() if d is not None else x

            # Check for Pseudo STANOX marker before stripping it
            note = 'Pseudo STANOX' if '*' in stanox else ''
            stanox = stanox.rstrip('*')

            # Extract notes within brackets/quotes
            n = re.search(r'(?<=[\[(\'])[\w, ]+.(?=[)\]\'])', x)

            if n is not None:
                note = '; '.join(item for item in [note, n.group()] if item != '')

            if '(' not in note and note.endswith(')'):
                note = note.rstrip(')')

    return stanox, note


def _parse_stanox_note(data):
    """
    Parses the note for STANOX.

    :param data: The preprocessed data of the location codes.
    :type data: pandas.DataFrame
    """

    col_name = 'STANOX'
    note_col_name = col_name + '_Note'

    # noinspection SpellCheckingInspection
    data[col_name] = data[col_name].str.replace('NANAN', '')

    if not data.empty:
        parsed_dat = data[col_name].map(_stanox_note).to_list()
        data[[col_name, note_col_name]] = pd.DataFrame(parsed_dat, index=data.index)
    else:
        # No data is available on the web page for the given 'key_word'
        data[note_col_name] = data[col_name]

    data[col_name] = data[col_name].str.replace('-', '')

    return data


def _fill_location_names(data):
    """
    Fills missing or empty code values based on other rows sharing the same 'Location' name.

    :param data: Input DataFrame with station data.
    :return: DataFrame with populated code columns.
    """

    df = data.copy()
    code_cols = ['CRS', 'NLC', 'TIPLOC', 'STANME', 'STANOX']  # The code columns to be filled

    # Standardise empty strings/None to NA so pandas recognises them as 'missing'
    for col in code_cols:
        df[col] = df[col].replace(r'^\s*$', pd.NA, regex=True)

    # Group and fill
    with pd.option_context('future.no_silent_downcasting', True):
        # Group by 'Location' and apply forward then backward fill
        df[code_cols] = df.groupby('Location', sort=False)[code_cols].transform(
            lambda x: x.ffill().bfill())

    # Replace any remaining NaNs back with empty strings (optional)
    df[code_cols] = df[code_cols].astype(object).fillna('')

    return df


def _parse_note_page_pre_span(pre_span):
    lines = [line.strip() for line in pre_span.decode_contents().splitlines() if line.strip()]

    data = []
    for line in lines:
        temp = bs4.BeautifulSoup(line, "html.parser")
        spans = temp.find_all('span')
        texts = [span.get_text(strip=True) for span in spans]

        # Remaining text after last span (e.g. "BLU", "EBF")
        remaining = temp.get_text(strip=True)
        for t in texts:
            remaining = remaining.replace(t, '', 1)
        remaining = remaining.strip()

        row = texts + ([remaining] if remaining else [])  # Build row: [Location, CRS1, CRS2, ..., ]
        data.append(row)

    if not data:
        return pd.DataFrame()

    # Determine dynamic column width
    max_len = max(len(row) for row in data)

    # Header: 'location_name' followed by incrementing CRS numbers
    columns = ['location_name'] + [f'CRS{i}' for i in range(1, max_len)]
    # Pad rows with empty strings to match max_len
    padded_data = [row + [''] * (max_len - len(row)) for row in data]

    return pd.DataFrame(padded_data, columns=columns)


def _format_structured_note(text):
    # noinspection PyShadowingNames
    """
    Formats a raw string of tab-separated notes into a structured DataFrame.

    This helper processes text blocks where data is organised by lines and delimited by tabs
    or multiple commas. It dynamically scales the number of 'CRS' columns based on the row
    with the most data points.

    :param text: Raw note text containing newline and tab characters.
    :type text: str
    :return: A structured DataFrame with columns ``['Location', 'CRS', 'CRS1', ...]``.
        Returns an empty DataFrame if required delimiters are not present.
    :rtype: pandas.DataFrame

    **Examples**::

        >>> from pyrcs.line_data.loc_id import _format_structured_note
        >>> text = "Abbey Wood\\tABW\\nAllerton\\tALN\\tLVP\\tLSP"
        >>> df = _format_structured_note(text)
        >>> df.columns.tolist()
        ['Location', 'CRS', 'CRS1']
    """

    if not text:
        return pd.DataFrame()

    # Normalise potential literal string escapes from web data
    # This ensures consistency across Windows/Unix and different scrape sources
    normalised_text = text.replace('\\t', '\t').replace('\\n', '\n').replace('\\r', '')

    if '\t' not in normalised_text:
        return pd.DataFrame()

    # Split into lines and remove empty strings
    lines = [line.strip() for line in normalised_text.splitlines() if line.strip()]

    processed_data = []
    max_cols = 0
    for line in lines:
        # Split by tabs or sequences of commas
        parts = [p.strip() for p in re.split(r'\t+|,+', line)]
        processed_data.append(parts)
        max_cols = max(max_cols, len(parts))

    if max_cols == 0:
        return pd.DataFrame()

    # Standardise columns: ['Location', 'CRS', 'CRS1', 'CRS2'...]
    columns = ['Location', 'CRS1']
    effective_max = max(2, max_cols)
    if effective_max > 2:
        columns += [f'CRS{i}' for i in range(2, effective_max)]

    # Match data to column length
    final_data = [row + [''] * (len(columns) - len(row)) for row in processed_data]

    return pd.DataFrame(final_data, columns=columns[:max_cols])


class LocationIdentifiers(_Base):
    """
    A class for collecting data of location identifiers
    (including `CRS, NLC, TIPLOC and STANOX codes <http://www.railwaycodes.org.uk/crs/crs0.shtm>`_
    and `other systems' station codes <http://www.railwaycodes.org.uk/crs/crs1.shtm>`_).

    The class retrieves and organises location identifiers for various railway stations and
    other related systems.
    """

    #: The name of the data.
    NAME: str = 'CRS, NLC, TIPLOC and STANOX codes'
    #: The key for accessing the data.
    KEY: str = 'Location ID'

    #: The URL of the main web page for the data.
    URL: str = urllib.parse.urljoin(homepage_url(), '/crs/crs0.shtm')

    #: The key for accessing the data of *other systems*.
    KEY_TO_OTHER_SYSTEMS: str = 'Other systems'
    #: The key for accessing the data of *multiple station codes explanatory note*.
    KEY_TO_MSCEN: str = 'Multiple station codes explanatory note'
    #: The key for accessing the data of *additional notes*.
    KEY_TO_NOTES: str = 'Notes'

    #: The key used to reference the last updated date in the data.
    KEY_TO_LAST_UPDATED_DATE: str = 'Last updated date'

    def __init__(self, data_dir=None, update=False, verbose=True):
        """
        :param data_dir: The name of the directory for storing the data; defaults to ``None``.
        :type data_dir: str | None
        :param update: Whether to check for updates to the data catalogue; defaults to ``False``.
        :type update: bool
        :param verbose: Whether to print relevant information to the console; defaults to ``True``.
        :type verbose: bool | int

        :ivar dict catalogue: The catalogue of the data.
        :ivar str last_updated_date: The date when the data was last updated.
        :ivar str data_dir: The path to the directory containing the data.
        :ivar str current_data_dir: The path to the current data directory.

        **Examples**::

            >>> from pyrcs.line_data import LocationIdentifiers
            >>> # from pyrcs import LocationIdentifiers
            >>> lid = LocationIdentifiers()
            >>> lid.NAME
            'CRS, NLC, TIPLOC and STANOX codes'
            >>> lid.URL
            'http://www.railwaycodes.org.uk/crs/crs0.shtm'
        """

        # data_cluster = re.sub(r",| codes| and", "", self.NAME.lower()).replace(" ", "-")
        super().__init__(
            data_dir=data_dir, content_type='catalogue', data_category="line-data",
            data_cluster="crs-nlc-tiploc-stanox", update=update, verbose=verbose)

        # Adds the multiple station codes explanatory note (MSCEN) to the catalogue
        mscen_url = urllib.parse.urljoin(homepage_url(), '/crs/crs2.shtm')
        self.catalogue.update({self.KEY_TO_MSCEN: mscen_url})

        # Retrieve the catalogue for other systems' station codes
        other_systems_url = self.catalogue[self.KEY_TO_OTHER_SYSTEMS]
        self.other_systems_catalogue = get_page_catalogue(url=other_systems_url)

    @staticmethod
    def _parse_notes_page(source, parser='html.parser'):
        """
        Parses the additional note page at the specified URL and extracts its contents.

        :param parser: The `parser`_ to use with `bs4.BeautifulSoup`_
            (e.g. ``'html.parser'``, ``'lxml'``); defaults to ``'html.parser'``.
        :type parser: str
        :return: A list of parsed text elements from the web page.
        :rtype: list

        .. _`parser`:
            https://www.crummy.com/software/BeautifulSoup/bs4/doc/
            index.html#specifying-the-parser-to-use
        .. _`bs4.BeautifulSoup`:
            https://www.crummy.com/software/BeautifulSoup/bs4/doc/index.html

        **Examples**::

            >>> from pyrcs.line_data import LocationIdentifiers
            >>> from pyhelpers.ops import fake_requests_headers
            >>> import requests
            >>> # from pyrcs import LocationIdentifiers
            >>> lid = LocationIdentifiers()
            >>> url = 'http://www.railwaycodes.org.uk/crs/crs2.shtm'
            >>> response = requests.get(url, headers=fake_requests_headers())
            >>> parsed_note_dat, _ = lid._parse_notes_page(response)
            >>> parsed_note_dat[3]
                          location_name CRS1 CRS2 CRS3
            0                 Bletchley  BLY  BLU
            1   Ebbsfleet International  EBD  EBF
            2           Glasgow Central  GLC  GCL
            3      Glasgow Queen Street  GLQ  GQL
            4                   Heworth  HEW  HEZ
            5      Highbury & Islington  HHY  HII  XHZ
            6    Lichfield Trent Valley  LTV  LIF
            7     Liverpool Lime Street  LIV  LVL
            8   Liverpool South Parkway  LPY  ALE
            9         London St Pancras  STP  SPL  SPX
            10                  Retford  RET  XRO
            11                 Tamworth  TAM  TAH
            12       Willesden Junction  WIJ  WJH  WJL
            13   Worcestershire Parkway  WOP  WPH
        """

        if not source or not source.ok:
            return [], None

        soup = bs4.BeautifulSoup(markup=source.content, features=parser)
        raw_elements = []

        # Extract elements from the page
        for x in soup.find_all(['p', 'pre']):
            if x.name == 'pre':
                raw_elements.append(_parse_note_page_pre_span(x))
            else:  # Check if the paragraph has direct string content
                text = x.get_text().strip().replace('  ', ' ')
                if text:
                    raw_elements.append(text)

        # Process elements into notes
        notes = []
        to_remove = {'click the link', 'click your browser', 'thank you', 'shown below'}

        for x in raw_elements:
            if isinstance(x, pd.DataFrame):
                notes.append(x)
            elif isinstance(x, str):
                if '\t' in x or '\\t' in x:  # If it looks like structured tab data
                    notes.append(_format_structured_note(x))
                else:  # Filter out boilerplate strings
                    if not any(phrase in x.lower() for phrase in to_remove):
                        notes.append(x)

        return notes, soup

    # -- CRS, NLC, TIPLOC and STANOX ---------------------------------------------------------------

    def _parse_crs_notes(self, data, initial, soup):
        # Identify rows that actually need a note lookup
        mask = data['CRS_Note'].str.contains('see note', case=False, na=False)
        if not mask.any():
            return None

        indices = data.index[mask].tolist()

        # Extract the specific 'note' links from the soup
        note_links = soup.find_all('a', href=True, string=re.compile(r'note', re.I))

        loc_id_notes = {}
        with requests.Session() as session:  # Using a Session is faster for multiple requests
            session.headers.update(fake_requests_headers())

            for idx, link_tag in zip(indices, note_links):
                crs_code = data.at[idx, 'CRS']
                # noinspection PyBroadException
                try:
                    url = urllib.parse.urljoin(self.catalogue[initial], link_tag['href'])
                    response = session.get(url, timeout=10)

                    parsed_content, _ = self._parse_notes_page(response)
                    # Get the first element if it's a list
                    loc_id_notes[crs_code] = parsed_content[0] if parsed_content else None

                except Exception:
                    loc_id_notes[crs_code] = None

        return loc_id_notes

    def _collect_loc_id(self, initial, source, verbose=False):
        initial_ = validate_initial(initial=initial)

        # url = lid.catalogue[initial_]
        # source = requests.get(url)
        soup = bs4.BeautifulSoup(markup=source.content, features='html.parser')

        thead, tbody = soup.find('thead'), soup.find('tbody')
        ths, trs = [th.get_text(strip=True) for th in thead.find_all('th')], tbody.find_all('tr')

        dat = parse_tr(trs=trs, ths=ths, sep=None, as_dataframe=True)
        dat = dat.replace({'\b-\b': '', '\xa0\xa0': ' ', '&half;': ' and 1/2'}, regex=True)

        data = dat.replace({'\xa0': ''}, regex=True)

        # Parse location names and their corresponding notes
        data[['Location', 'Location_Note']] = pd.DataFrame(  # Collect additional info as note
            data['Location'].map(_parse_raw_location_name).to_list())
        # # Debugging
        # for i, x in enumerate(data['Location']):
        #     try:
        #         _parse_raw_location_name(x)
        #     except Exception:
        #         print(i)
        #         break
        data.replace(_amendment_to_location_names(), regex=True, inplace=True)

        # Cleanse multiple alternatives for every code column
        data = _parse_mult_alt_codes(data=data)

        # Parse note for every code column
        data = _parse_code_notes(data=data)

        # Parse STANOX note
        data = _parse_stanox_note(data=data)

        # Fills missing or empty code values based on other rows sharing the same 'Location' name.
        data = _fill_location_names(data)

        loc_codes = {
            initial_: data,
            self.KEY_TO_NOTES: self._parse_crs_notes(data=data, initial=initial_, soup=soup),
            self.KEY_TO_LAST_UPDATED_DATE: _get_last_updated_date(soup=soup),
        }

        if verbose in {True, 1}:
            print("Done.")

        self._save_data_to_file(data=loc_codes, data_name=initial_, sub_dir="a-z", verbose=verbose)

        return loc_codes

    def collect_loc_id(self, initial, confirmation_required=True, verbose=False, raise_error=False):
        """
        Collects `CRS, NLC, TIPLOC, STANME and STANOX codes
        <http://www.railwaycodes.org.uk/crs/crs0.shtm>`_ for a given initial letter.

        :param initial: The initial letter (e.g. ``'a'``, ``'z'``) of a location name.
        :type initial: str
        :param confirmation_required: Whether user confirmation is required;
            if ``confirmation_required=True`` (default), prompts the user for confirmation
            before proceeding with data collection.
        :type confirmation_required: bool
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :param raise_error: Whether to raise the provided exception;
            if ``raise_error=False`` (default), the error will be suppressed.
        :type raise_error: bool
        :return: A dictionary containing data of locations whose names start with the given
            initial letter, along with the date of the last update.
        :rtype: dict

        **Examples**::

            >>> from pyrcs.line_data import LocationIdentifiers
            >>> # from pyrcs import LocationIdentifiers
            >>> lid = LocationIdentifiers()
            >>> loc_a_codes = lid.collect_loc_id(initial='a')
            To collect data of CRS, NLC, TIPLOC and STANOX codes beginning with "A"
            ? [No]|Yes: yes
            >>> type(loc_a_codes)
            dict
            >>> list(loc_a_codes.keys())
            ['A', 'Additional notes', 'Last updated date']
            >>> loc_a_codes_dat = loc_a_codes['A']
            >>> type(loc_a_codes_dat)
            pandas.core.frame.DataFrame
            >>> loc_a_codes_dat.head()
                                          Location CRS  ... STANME_Note STANOX_Note
            0                 1999 Reorganisations      ...
            1                                   A1      ...
            2                       A463 Traded In      ...
            3  A483 Road Scheme Supervisors Closed      ...
            4                               Aachen      ...
            [5 rows x 12 columns]
        """

        initial_ = validate_initial(initial=initial)

        loc_id_data = self._collect_data_from_source(
            data_name=self.NAME, method=self._collect_loc_id, initial=initial_,
            additional_fields=self.KEY_TO_NOTES, confirmation_required=confirmation_required,
            verbose=verbose, raise_error=raise_error)

        return loc_id_data

    def fetch_loc_id(self, initial=None, update=False, dump_dir=None, verbose=False, **kwargs):
        """
        Fetches data of `CRS, NLC, TIPLOC, STANME and STANOX codes`_.

        .. _`CRS, NLC, TIPLOC, STANME and STANOX codes`:
            http://www.railwaycodes.org.uk/crs/crs0.shtm

        :param initial: The initial letter (e.g. ``'a'``, ``'z'``) of a location name.
        :type initial: str
        :param update: Whether to check for updates to the package data; defaults to ``False``.
        :type update: bool
        :param dump_dir: Path to a directory where the data file will be saved;
            defaults to ``None``.
        :type dump_dir: str | None
        :param verbose: Whether to print relevant information to the console; defaults to ``True``.
        :type verbose: bool | int
        :return: A dictionary containing data of locations whose names start with the given
            initial letter, along with the date of the last update.
        :rtype: dict

        **Examples**::

            >>> from pyrcs.line_data import LocationIdentifiers
            >>> # from pyrcs import LocationIdentifiers
            >>> lid = LocationIdentifiers()
            >>> loc_a_codes = lid.fetch_loc_id(initial='a')
            >>> type(loc_a_codes)
            dict
            >>> list(loc_a_codes.keys())
            ['A', 'Additional notes', 'Last updated date']
            >>> loc_a_codes_dat = loc_a_codes['A']
            >>> type(loc_a_codes_dat)
            pandas.core.frame.DataFrame
            >>> loc_a_codes_dat.head()
                                          Location CRS  ... STANME_Note STANOX_Note
            0                 1999 Reorganisations      ...
            1                                   A1      ...
            2                       A463 Traded In      ...
            3  A483 Road Scheme Supervisors Closed      ...
            4                               Aachen      ...
            [5 rows x 12 columns]
            >>> loc_codes = lid.fetch_loc_id()
            >>> list(loc_codes.keys())
            ['Location ID', 'Last updated date']
            >>> loc_codes[lid.KEY].shape
            (59873, 12)
        """

        if initial:
            args = {
                'data_name': validate_initial(initial),
                'method': self.collect_loc_id,
                'sub_dir': "a-z",
                'initial': initial,
            }
            kwargs.update(args)

            loc_id_data = self._fetch_data_from_file(
                update=update, dump_dir=dump_dir, verbose=verbose, **kwargs)

        else:
            verbose_1 = get_collect_verbosity_for_fetch(data_dir=dump_dir, verbose=verbose)

            # Get every data table
            verbose_2 = verbose_1 if is_homepage_connectable() else False
            dat_list = [
                self.fetch_loc_id(initial=x, update=update, verbose=verbose_2)
                for x in string.ascii_lowercase]

            if all(d[x] is None for d, x in zip(dat_list, string.ascii_uppercase)):
                if update:
                    print_instance_connection_error(verbose=verbose)
                    print_void_collection_message(data_name=self.KEY, verbose=verbose)

                dat_list = [
                    self.fetch_loc_id(initial=x, update=False, verbose=verbose_1)
                    for x in string.ascii_lowercase]

            # Select DataFrames only
            data = pd.concat(
                (item[x] for item, x in zip(dat_list, string.ascii_uppercase)), ignore_index=True)

            # Get the latest updated date
            last_updated_dates = (
                item[self.KEY_TO_LAST_UPDATED_DATE]
                for item, _ in zip(dat_list, string.ascii_uppercase))
            latest_update_date = max(d for d in last_updated_dates if d is not None)

            loc_id_data = {self.KEY: data, self.KEY_TO_LAST_UPDATED_DATE: latest_update_date}

        return loc_id_data

    # -- Other systems -----------------------------------------------------------------------------

    @staticmethod
    def _parse_code(x):
        protocol = 'https://'

        if '; ' in x and protocol in x:
            temp = x.split('; ')
            x0, x1 = temp[0], [y.split(protocol) for y in temp if protocol in y][0]
            x1 = ' ('.join([x1[0], protocol + x1[1]]) + ')'
            x_ = [x0, x1]
        else:
            x_ = [x, '']

        return x_

    def _parse_tbl_dat(self, h3_or_h4):
        tbl_dat = h3_or_h4.find_next('thead'), h3_or_h4.find_next('tbody')
        thead, tbody = tbl_dat
        ths = [x.text for x in thead.find_all('th')]
        trs = tbody.find_all('tr')
        tbl = parse_tr(trs=trs, ths=ths, as_dataframe=True)

        if 'Code' in tbl.columns:
            if tbl.Code.str.contains('https://').sum() > 0:
                temp = tbl['Code'].map(self._parse_code)
                tbl_ext = pd.DataFrame(zip(*temp)).T
                tbl_ext.columns = ['Code', 'Code_extra']
                del tbl['Code']
                tbl = pd.concat([tbl, tbl_ext], axis=1, sort=False)

        return tbl

    def _collect_other_systems_codes(self, source, verbose=False):
        soup = bs4.BeautifulSoup(markup=source.content, features='html.parser')

        other_systems_codes = collections.defaultdict(dict)

        for h3 in soup.find_all('h3'):
            h4 = h3.find_next('h4')

            if h4 is not None:
                while h4:
                    prev_h3 = h4.find_previous('h3')
                    if prev_h3.text == h3.text:
                        other_systems_codes[h3.text].update(
                            {h4.text: self._parse_tbl_dat(h4)})
                        h4 = h4.find_next('h4')
                    elif h3.text not in other_systems_codes.keys():
                        other_systems_codes.update({h3.text: self._parse_tbl_dat(h3)})
                        break
                    else:
                        break
            else:
                other_systems_codes.update({h3.text: self._parse_tbl_dat(h3)})

        other_systems_codes = {
            self.KEY_TO_OTHER_SYSTEMS: dict(other_systems_codes),
            self.KEY_TO_LAST_UPDATED_DATE: _get_last_updated_date(soup=soup),
        }

        if verbose in {True, 1}:
            print("Done.")

        self._save_data_to_file(
            data=other_systems_codes, data_name=self.KEY_TO_OTHER_SYSTEMS, verbose=verbose)

        return other_systems_codes

    def collect_other_systems_codes(self, confirmation_required=True, verbose=False,
                                    raise_error=False):
        """
        Collects data of `other systems' station codes`_ from the source web page.

        .. _`other systems' station codes`: http://www.railwaycodes.org.uk/crs/crs1.shtm

        :param confirmation_required: Whether user confirmation is required before proceeding;
            defaults to ``True``.
        :type confirmation_required: bool
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :param raise_error: Whether to raise the provided exception; defaults to ``True``.
            if ``raise_error=False``, the error will be suppressed.
        :type raise_error: bool
        :return: A dictionary containing station codes for other systems,
            or ``None`` if no data is collected.
        :rtype: dict | None

        **Examples**::

            >>> from pyrcs.line_data import LocationIdentifiers
            >>> # from pyrcs import LocationIdentifiers
            >>> lid = LocationIdentifiers()
            >>> os_codes = lid.collect_other_systems_codes()
            To collect data of Other systems
            ? [No]|Yes: yes
            >>> type(os_codes)
            dict
            >>> list(os_codes.keys())
            ['Other systems', 'Last updated date']
            >>> lid.KEY_TO_OTHER_SYSTEMS
            'Other systems'
            >>> os_codes_dat = os_codes[lid.KEY_TO_OTHER_SYSTEMS]
            >>> type(os_codes_dat)
            dict
            >>> list(os_codes_dat.keys())
            ['Córas Iompair Éireann (Republic of Ireland)',
             'Crossrail',
             'Croydon Tramlink',
             'Docklands Light Railway',
             'Manchester Metrolink',
             'Translink (Northern Ireland)',
             'Tyne & Wear Metro']
        """

        other_systems_codes = self._collect_data_from_source(
            data_name=self.KEY_TO_OTHER_SYSTEMS.lower(), method=self._collect_other_systems_codes,
            url=self.catalogue[self.KEY_TO_OTHER_SYSTEMS],
            confirmation_required=confirmation_required, verbose=verbose, raise_error=raise_error)

        return other_systems_codes

    def fetch_other_systems_codes(self, update=False, dump_dir=None, verbose=False, **kwargs):
        """
        Fetches data of `other systems' station codes`_.

        .. _`other systems' station codes`: http://www.railwaycodes.org.uk/crs/crs1.shtm

        :param update: Whether to check for updates to the package data; defaults to ``False``.
        :type update: bool
        :param dump_dir: The path to a directory where the data file will be saved;
            defaults to ``None``.
        :type dump_dir: str | None
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :return: A dictionary containing station codes for other systems.
        :rtype: dict

        **Examples**::

            >>> from pyrcs.line_data import LocationIdentifiers
            >>> # from pyrcs import LocationIdentifiers
            >>> lid = LocationIdentifiers()
            >>> os_codes = lid.fetch_other_systems_codes()
            >>> type(os_codes)
            dict
            >>> list(os_codes.keys())
            ['Other systems', 'Last updated date']
            >>> lid.KEY_TO_OTHER_SYSTEMS
            'Other systems'
            >>> os_codes_dat = os_codes[lid.KEY_TO_OTHER_SYSTEMS]
            >>> type(os_codes_dat)
            collections.defaultdict
            >>> list(os_codes_dat.keys())
            ['Córas Iompair Éireann (Republic of Ireland)',
             'Crossrail',
             'Croydon Tramlink',
             'Docklands Light Railway',
             'Manchester Metrolink',
             'Translink (Northern Ireland)',
             'Tyne & Wear Metro']
        """

        kwargs.update(
            {'data_name': self.KEY_TO_OTHER_SYSTEMS, 'method': self.collect_other_systems_codes})

        other_systems_codes = self._fetch_data_from_file(
            update=update, dump_dir=dump_dir, verbose=verbose, **kwargs)

        return other_systems_codes

    # -- Additional notes --------------------------------------------------------------------------

    def _collect_notes(self, source, verbose=False, parser='html.parser'):
        notes_dat, soup = self._parse_notes_page(source=source, parser=parser)

        notes = {
            self.KEY_TO_NOTES: {self.KEY_TO_MSCEN: [x for x in notes_dat if 'Last update' not in x]},
            self.KEY_TO_LAST_UPDATED_DATE: _get_last_updated_date(soup=soup),
        }

        if verbose in {True, 1}:
            print("Done.")

        self._save_data_to_file(data=notes, data_name=self.KEY_TO_NOTES, verbose=verbose)

        return notes

    def collect_notes(self, confirmation_required=True, verbose=False, raise_error=False):
        # noinspection PyShadowingNames
        """
        Collects the explanatory note related to multiple station codes (CRS codes)
        from the source web page.

        :param confirmation_required: Whether user confirmation is required;
            if ``confirmation_required=True`` (default), prompts the user for confirmation
            before proceeding with data collection.
        :type confirmation_required: bool
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :param raise_error: Whether to raise the provided exception;
            if ``raise_error=False`` (default), the error will be suppressed.
        :type raise_error: bool
        :return: A dictionary containing the data of the multiple station codes explanatory note,
            or ``None`` if the note is not found or the collection is not performed.
        :rtype: dict | None

        **Examples**::

            >>> from pyrcs.line_data import LocationIdentifiers
            >>> # from pyrcs import LocationIdentifiers
            >>> lid = LocationIdentifiers()
            >>> notes = lid.collect_notes()
            To collect data of multiple station codes explanatory note
            ? [No]|Yes: yes
            >>> type(notes)
            dict
            >>> list(notes.keys())
            ['Notes', 'Last updated date']
            >>> lid.KEY_TO_NOTES
            'Notes'
            >>> notes_ = notes[lid.KEY_TO_NOTES]
            >>> type(notes_)
            dict
            >>> notes_[lid.KEY_TO_MSCEN][2].head()
                         location_name CRS1 CRS2 CRS3
            0                Bletchley  BLY  BLU
            1  Ebbsfleet International  EBD  EBF
            2          Glasgow Central  GLC  GCL
            3     Glasgow Queen Street  GLQ  GQL
            4                  Heworth  HEW  HEZ
        """

        explanatory_notes = self._collect_data_from_source(
            data_name="additional notes", method=self._collect_notes,
            url=self.catalogue[self.KEY_TO_MSCEN],
            confirmation_required=confirmation_required,
            confirmation_prompt=format_confirmation_prompt(data_name="additional notes"),
            verbose=verbose, raise_error=raise_error)

        return explanatory_notes

    def fetch_notes(self, update=False, dump_dir=None, verbose=False, **kwargs):
        """
        Fetches the explanatory note for multiple station codes (CRS codes).

        :param update: Whether to check for updates to the package data; defaults to ``False``.
        :type update: bool
        :param dump_dir: The path to a directory where the data file will be saved;
            defaults to ``None``.
        :type dump_dir: str | None
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :return: A dictionary containing the data of the multiple station codes explanatory note.
        :rtype: dict

        **Examples**::

            >>> from pyrcs.line_data import LocationIdentifiers
            >>> # from pyrcs import LocationIdentifiers
            >>> lid = LocationIdentifiers()
            >>> exp_note = lid.fetch_notes()
            >>> type(exp_note)
            dict
            >>> list(exp_note.keys())
            ['Multiple station codes explanatory note', 'Notes', 'Last updated date']
            >>> lid.KEY_TO_MSCEN
            'Multiple station codes explanatory note'
            >>> exp_note_dat = exp_note[lid.KEY_TO_MSCEN]
            >>> type(exp_note_dat)
            pandas.core.frame.DataFrame
            >>> exp_note_dat.head()
                             Location  CRS CRS_alt1 CRS_alt2
            0         Glasgow Central  GLC      GCL
            1    Glasgow Queen Street  GLQ      GQL
            2                 Heworth  HEW      HEZ
            3    Highbury & Islington  HHY      HII      XHZ
            4  Lichfield Trent Valley  LTV      LIF
        """

        kwargs.update({'data_name': self.KEY_TO_NOTES, 'method': self.collect_notes})

        explanatory_note = self._fetch_data_from_file(
            update=update, dump_dir=dump_dir, verbose=verbose, **kwargs)

        return explanatory_note

    # -- All codes ---------------------------------------------------------------------------------

    def fetch_codes(self, update=False, dump_dir=None, verbose=False, **kwargs):
        """
        Fetches location codes listed in the `CRS, NLC, TIPLOC and STANOX codes`_ catalogue
        (including `other systems' station codes`_).

        .. _`CRS, NLC, TIPLOC and STANOX codes`:
            http://www.railwaycodes.org.uk/crs/crs0.shtm
        .. _`other systems' station codes`:
            http://www.railwaycodes.org.uk/crs/crs1.shtm

        :param update: Whether to check for updates to the package data; defaults to ``False``.
        :type update: bool
        :param dump_dir: The path to a directory where the data file will be saved;
            defaults to ``None``.
        :type dump_dir: str | None
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :return: A dictionary containing location codes and date of when the data was last updated.
        :rtype: dict

        **Examples**::

            >>> from pyrcs.line_data import LocationIdentifiers
            >>> # from pyrcs import LocationIdentifiers
            >>> lid = LocationIdentifiers()
            >>> loc_codes = lid.fetch_codes()
            >>> type(loc_codes)
            dict
            >>> list(loc_codes.keys())
            ['Location ID', 'Other systems', 'Additional notes', 'Last updated date']
            >>> lid.KEY
            'LocationID'
            >>> loc_codes_dat = loc_codes[lid.KEY]
            >>> type(loc_codes_dat)
            pandas.core.frame.DataFrame
            >>> loc_codes_dat.head()
                                          Location CRS  ... STANME_Note STANOX_Note
            0                 1999 Reorganisations      ...
            1                                   A1      ...
            2                       A463 Traded In      ...
            3  A483 Road Scheme Supervisors Closed      ...
            4                               Aachen      ...
            [5 rows x 12 columns]
        """

        verbose_ = get_collect_verbosity_for_fetch(data_dir=dump_dir, verbose=verbose)

        loc_id_data = self.fetch_loc_id(update=update, verbose=verbose_)

        # Get other systems codes
        other_systems_codes = self.fetch_other_systems_codes(update=update, verbose=verbose_)

        # Get explanatory notes
        explanatory_notes = self.fetch_notes(update=update, verbose=verbose_)

        keys = [self.KEY, self.KEY_TO_OTHER_SYSTEMS, self.KEY_TO_NOTES]
        data_list = [loc_id_data, other_systems_codes, explanatory_notes]

        location_codes = {}
        for dat, key in zip(data_list, keys):
            if dat is None:
                location_codes.update({key: None})
            else:
                for k, v in dat.items():
                    if k != self.KEY_TO_LAST_UPDATED_DATE:
                        location_codes.update({k: v})

        # Get the latest updated date
        latest_update_date = max(d[self.KEY_TO_LAST_UPDATED_DATE] for d in filter(None, data_list))

        # Create a dict to include all information
        location_codes.update({self.KEY_TO_LAST_UPDATED_DATE: latest_update_date})

        if dump_dir:
            kwargs.update({'data': location_codes, 'data_name': self.KEY})
            self._save_data_to_file(dump_dir=dump_dir, verbose=verbose, **kwargs)

        return location_codes

    @staticmethod
    def _make_xref_dict(location_codes, keys, drop_duplicates=False, as_dict=False, main_key=None):
        # Extract relevant columns and remove empty rows
        key_locid = location_codes[['Location'] + keys].query(
            ' | '.join([f"{k} != ''" for k in keys]))

        # Further clean location_code
        if drop_duplicates:
            location_codes_ref = key_locid.drop_duplicates(keys, keep='first').set_index(keys)

        else:  # drop_duplicates is False or None
            locid_unique = key_locid.drop_duplicates(subset=keys, keep=False)

            dupl_temp_1 = key_locid[key_locid.duplicated(['Location'] + keys, keep=False)]
            dupl_temp_2 = key_locid[key_locid.duplicated(keys, keep=False)]
            duplicated_1 = dupl_temp_2[dupl_temp_1.eq(dupl_temp_2)].dropna().drop_duplicates()
            duplicated_2 = dupl_temp_2[~dupl_temp_1.eq(dupl_temp_2)].dropna()
            duplicated_entries = pd.concat([duplicated_1, duplicated_2], axis=0)

            grouped_duplicates = duplicated_entries.groupby(keys).agg(tuple)
            grouped_duplicates['Location'] = grouped_duplicates['Location'].map(
                lambda x: x[0] if len(set(x)) == 1 else x)

            locid_unique.set_index(keys, inplace=True)
            location_codes_ref = pd.concat([locid_unique, grouped_duplicates])

        if as_dict:
            location_codes_dict = location_codes_ref.to_dict()
            if main_key:
                location_codes_dict = {main_key: location_codes_dict.pop('Location')}
            else:
                location_codes_dict = location_codes_dict['Location']

        else:
            location_codes_dict = location_codes_ref

        return location_codes_dict

    def make_xref_dict(self, keys, initials=None, drop_duplicates=False, as_dict=False,
                       main_key=None, update=False, dump_it=False, dump_dir=None, verbose=False):
        """
        Creates a dictionary or dataframe containing location code data for the specified ``keys``.

        :param keys: A string or list of keys from ``['CRS', 'NLC', 'TIPLOC', 'STANOX', 'STANME']``
            representing the location codes.
        :type keys: str | list
        :param initials: A string or list of initials for which the codes are filtered;
            defaults to ``None``.
        :type initials: str | list | None
        :param drop_duplicates: If ``True``, removes duplicate entries from the result;
            defaults to ``False``.
        :type drop_duplicates: bool
        :param as_dict: If ``True``, returns the result as a dictionary;
            otherwise, returns a dataframe; defaults to ``False``.
        :type as_dict: bool
        :param main_key: The key used for the returned dictionary when ``as_dict=True``;
            defaults to ``None``.
        :type main_key: str | None
        :param update: Whether to check for updates to the package data; defaults to ``False``.
        :type update: bool
        :param dump_it: If ``True``, saves the result to a file; defaults to ``False``.
        :type dump_it: bool
        :param dump_dir: The directory path where the file can be saved, if ``dump_it=True``;
            defaults to ``None``.
        :type dump_dir: str | None
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :return: A dictionary or dataframe containing cross-reference location data
            for the specified ``keys``, or ``None`` if no data is available.
        :rtype: dict | pandas.DataFrame | None

        **Examples**::

            >>> from pyrcs.line_data import LocationIdentifiers
            >>> # from pyrcs import LocationIdentifiers
            >>> lid = LocationIdentifiers()
            >>> stanox_dict = lid.make_xref_dict(keys='STANOX')
            >>> type(stanox_dict)
            pandas.core.frame.DataFrame
            >>> stanox_dict.head()
                                      Location
            STANOX
            00005                       Aachen
            04309           Abbeyhill Junction
            04311        Abbeyhill Signal E811
            04308   Abbeyhill Turnback Sidings
            88601                   Abbey Wood
            >>> s_t_dictionary = lid.make_xref_dict(keys=['STANOX', 'TIPLOC'], initials='a')
            >>> type(s_t_dictionary)
            pandas.core.frame.DataFrame
            >>> s_t_dictionary.head()
                                              Location
            STANOX TIPLOC
            00005  AACHEN                       Aachen
            04309  ABHLJN           Abbeyhill Junction
            04311  ABHL811       Abbeyhill Signal E811
            04308  ABHLTB   Abbeyhill Turnback Sidings
            88601  ABWD                     Abbey Wood
            >>> ks = ['STANOX', 'TIPLOC']
            >>> ini = 'b'
            >>> main_k = 'Data'
            >>> s_t_dictionary = lid.make_xref_dict(ks, ini, main_k, as_dict=True)
            >>> type(s_t_dictionary)
            dict
            >>> list(s_t_dictionary.keys())
            ['Data']
            >>> list(s_t_dictionary['Data'].keys())[:5]
            [('55115', ''),
             ('23490', 'BABWTHL'),
             ('38306', 'BACHE'),
             ('66021', 'BADESCL'),
             ('81003', 'BADMTN')]
        """

        valid_keys = {'CRS', 'NLC', 'TIPLOC', 'STANOX', 'STANME'}
        if isinstance(keys, str):
            keys = [keys]
        if not all(k in valid_keys for k in keys):
            raise ValueError(f"`keys` must be one of {valid_keys}, but got {keys}.")

        if main_key and not isinstance(main_key, str):
            raise TypeError("`main_key` must be a string.")

        # Validate `initials`
        if initials:
            if isinstance(initials, str):
                initials = [validate_initial(initials, as_is=True)]
            elif not all(isinstance(x, str) and x in string.ascii_letters for x in initials):
                raise ValueError("`initials` must be a string or a list of letters.")

            dat_list = [
                self.fetch_loc_id(x, update=update, verbose=verbose).get(x.upper())
                for x in initials]
            location_codes = pd.concat(dat_list, ignore_index=True)

        else:
            location_codes = self.fetch_codes(update=update, verbose=verbose).get(self.KEY)

        if verbose == 2:
            print("Generating location code dictionary", end=" ... ")

        try:
            location_codes_dict = self._make_xref_dict(
                location_codes, keys=keys, drop_duplicates=drop_duplicates, as_dict=as_dict,
                main_key=main_key)

            if verbose == 2:
                print("Done.")

            if dump_it:
                self._save_data_to_file(
                    data=location_codes_dict,
                    data_name="-".join(keys) + (f"-{''.join(initials)}" if initials else ""),
                    ext=".json" if as_dict and len(keys) == 1 else ".pkl",
                    dump_dir=validate_dir(dump_dir) if dump_dir else self._cdd("xref-dicts"),
                    verbose=verbose)

            return location_codes_dict

        except Exception as e:
            _print_failure_message(e)



"""
Collects data of `railway line names <http://www.railwaycodes.org.uk/misc/line_names.shtm>`_.
"""

import re
import urllib.parse

import pandas as pd

from .._base import _Base
from ..parser import _get_last_updated_date, parse_table
from ..utils import homepage_url


class LineNames(_Base):
    """
    A class for collecting data of
    `railway line names <http://www.railwaycodes.org.uk/misc/line_names.shtm>`_.
    """

    #: The name of the data.
    NAME: str = 'Railway line names'
    #: The key for accessing the data.
    KEY: str = 'Line names'

    #: The URL of the main web page for the data.
    URL: str = urllib.parse.urljoin(homepage_url(), '/line/line_names.shtm')

    #: The key used to reference the last updated date in the data.
    KEY_TO_LAST_UPDATED_DATE: str = 'Last updated date'

    def __init__(self, data_dir=None, update=False, verbose=True):
        """
        :param data_dir: The name of the directory for storing the data; defaults to ``None``.
        :type data_dir: str | None
        :param update: Whether to check for updates to the data catalogue; defaults to ``False``.
        :type update: bool
        :param verbose: Whether to print relevant information to the console; defaults to ``True``.
        :type verbose: bool | int

        :ivar dict catalogue: The catalogue of the data.
        :ivar str last_updated_date: The date when the data was last updated.
        :ivar str data_dir: The path to the directory containing the data.
        :ivar str current_data_dir: The path to the current data directory.

        **Examples**::

            >>> from pyrcs.line_data import LineNames  # from pyrcs import LineNames
            >>> ln = LineNames()
            >>> ln.NAME
            'Railway line names'
            >>> ln.URL
            'http://www.railwaycodes.org.uk/misc/line_names.shtm'
        """

        super().__init__(
            data_dir=data_dir, content_type='catalogue', data_category="line-data", update=update,
            verbose=verbose)

    @staticmethod
    def _parse_route(x):
        """Parse route column."""
        if 'Watford - Euston suburban route' in x:
            route, route_note = 'Watford - Euston suburban route', x

        elif ', including Moorgate - Farringdon' in x:
            route_note = 'including Moorgate - Farringdon'
            route = x.replace(', including Moorgate - Farringdon', '')

        elif re.match(r'.+(?= \[\')', x):
            route, route_note = re.split(r' \[\'\(?', x)
            route_note = route_note.strip(")']")

        elif re.match(r'.+\)$', x):
            if re.match(r'.+(?= - \()', x):
                route, route_note = x, None
            else:
                route, route_note = re.split(r' \(\[?\'?', x)
                route_note = route_note.rstrip('\'])')

        else:
            route, route_note = x, None

        return route, route_note

    def _collect_codes(self, source, verbose=False):
        (columns, records), soup = parse_table(source=source)
        line_names = pd.DataFrame(
            [[rec.replace('\xa0', '').strip() for rec in record] for record in records],
            columns=columns)

        rte_col = ['Route', 'Route_note']
        temp = line_names['Route'].map(self._parse_route)
        line_names[rte_col] = pd.DataFrame(zip(*temp)).T

        last_updated_date = _get_last_updated_date(soup=soup)
        line_names_data = {self.KEY: line_names, self.KEY_TO_LAST_UPDATED_DATE: last_updated_date}

        if verbose in {True, 1}:
            print("Done.")

        self._save_data_to_file(data=line_names_data, data_name=self.KEY, verbose=verbose)

        return line_names_data

    def collect_codes(self, confirmation_required=True, verbose=False, raise_error=False):
        """
        Collects data of `railway line names`_ and associated route data from the source web page.

        .. _`railway line names`: http://www.railwaycodes.org.uk/misc/line_names.shtm

        :param confirmation_required: Whether user confirmation is required;
            if ``confirmation_required=True`` (default), prompts the user for confirmation
            before proceeding with data collection.
        :type confirmation_required: bool
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :param raise_error: Whether to raise the provided exception;
            if ``raise_error=False`` (default), the error will be suppressed.
        :type raise_error: bool
        :return: A dictionary containing railway line names, route data and the last update date,
            or ``None`` if no data is collected.
        :rtype: dict | None

        **Examples**::

            >>> from pyrcs.line_data import LineNames  # from pyrcs import LineNames
            >>> ln = LineNames()
            >>> line_names_codes = ln.collect_codes()
            To collect British railway line names
            ? [No]|Yes: yes
            >>> type(line_names_codes)
            dict
            >>> list(line_names_codes.keys())
            ['Line names', 'Last updated date']
            >>> ln.KEY
            'Line names'
            >>> line_names_codes_dat = line_names_codes[ln.KEY]
            >>> type(line_names_codes_dat)
            pandas.core.frame.DataFrame
            >>> line_names_codes_dat.head()
                      Line name  ... Route_note
            0        Abbey Line  ...       None
            1     Aberdare Line  ...       None
            2     Airedale Line  ...       None
            3       Argyle Line  ...       None
            4  Arun Valley Line  ...       None
            [5 rows x 3 columns]
        """

        data = self._collect_data_from_source(
            data_name=self.NAME.lower(), method=self._collect_codes, url=self.URL,
            confirmation_required=confirmation_required, verbose=verbose,
            raise_error=raise_error)

        return data

    def fetch_codes(self, update=False, dump_dir=None, verbose=False, **kwargs):
        """
        Fetches data of `railway line names`_ and associated route data.

        .. _`railway line names`: http://www.railwaycodes.org.uk/misc/line_names.shtm

        :param update: Whether to check for updates to the package data; defaults to ``False``.
        :type update: bool
        :param dump_dir: The path to a directory where the data file will be saved;
            defaults to ``None``.
        :type dump_dir: str | None
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :return: A dictionary containing railway line names, route data and the last update date.
        :rtype: dict

        **Examples**::

            >>> from pyrcs.line_data import LineNames  # from pyrcs import LineNames
            >>> ln = LineNames()
            >>> line_names_codes = ln.fetch_codes()
            >>> type(line_names_codes)
            dict
            >>> list(line_names_codes.keys())
            ['Line names', 'Last updated date']
            >>> ln.KEY
            'Line names'
            >>> line_names_codes_dat = line_names_codes[ln.KEY]
            >>> type(line_names_codes_dat)
            pandas.core.frame.DataFrame
            >>> line_names_codes_dat.head()
                         Line name  ... Route_note
            0           Abbey Line  ...       None
            1        Airedale Line  ...       None
            2          Argyle Line  ...       None
            3     Arun Valley Line  ...       None
            4  Atlantic Coast Line  ...       None
            [5 rows x 3 columns]
        """

        kwargs.update({'data_name': self.KEY, 'method': self.collect_codes})

        line_names_data = self._fetch_data_from_file(
            update=update, dump_dir=dump_dir, verbose=verbose, **kwargs)

        return line_names_data
